







-- =============================================
-- Author:	PMX
-- Create date: 
-- Description:	Produmex extension on the SBO_SP_TransactionNotification for Dirk Raes
-- =============================================
CREATE PROCEDURE [dbo].[PMX_SP_Troy_TransactionNotification] 
	-- Add the parameters for the stored procedure here
	@object_type nvarchar(20), 			-- SBO Object Type
	@transaction_type nchar(1),			-- [A]dd, [U]pdate, [D]elete, [C]ancel, C[L]ose
	@num_of_cols_in_key int,
	@list_of_key_cols_tab_del nvarchar(255),
	@list_of_cols_val_tab_del nvarchar(255),
	@error int OUTPUT, -- Result (0 for no error)
	@error_message nvarchar (200) OUTPUT -- Error string to be displayed

AS

BEGIN
	--**********************************************************************************************************
	--Return and handling errors:
	--If a problem is found the function sets the parameters @error and @error_message
	--and logs the error in the windows event logging
	--**********************************************************************************************************
	
	--**********************************************************************************************************
	--Logging errors:
	--If there are errors and this procedure is called when using the DIAPI then the user does not see the error
	--Because of this problem the errors are also logged in the windows event logging
	--To log you have to use the funciton: EXEC xp_logevent @error, @error_message, ERROR
	--**********************************************************************************************************

	--Logging
	DECLARE @MSG NVARCHAR (255)
	DECLARE @errorList AS NVARCHAR(255)
	DECLARE @count AS INT
	DECLARE @ForceBatch AS VARCHAR(1)
	DECLARE @SeriesName AS VARCHAR (10)

	--Set default values
	SET @error = 0
	SET @error_message = ''
	SET @SeriesName = ''

	--At this moment the function supports: purchase delivery note, sales delivery note, move, item and BOM
	DECLARE @isPurchaseDeliveryDocument AS BIT
	DECLARE @isPurchaseReturnDocument AS BIT
	DECLARE @isSalesDeliveryDocument AS BIT
	DECLARE @isSalesReturnDocument AS BIT
	DECLARE @isSalesOrderDocument AS BIT
	DECLARE @isInventoryEntry AS BIT
	DECLARE @isInventoryExit AS BIT
	DECLARE @isMoveDoc AS BIT
	DECLARE @isItem AS BIT
	DECLARE @isTransactionTypeAdd AS BIT
	DECLARE @isTransactionTypeUpdate AS BIT
	DECLARE @isTransactionTypeDelete AS BIT
	DECLARE @isBOM AS BIT
	DECLARE @isUser AS BIT
	DECLARE @isDocument AS BIT
	DECLARE @isPurchaseDocument AS BIT
    DECLARE @isSalesCreditNote AS BIT
	DECLARE @isSalesInvoice AS BIT
	DECLARE @IsPicklist AS BIT
	
	SET @isPurchaseDeliveryDocument = 0
	SET @isPurchaseReturnDocument = 0
	SET @isSalesDeliveryDocument = 0
	SET @isSalesReturnDocument = 0
	SET @isSalesOrderDocument = 0
	SET @isInventoryEntry = 0
	SET @isInventoryExit = 0
	SET @isMoveDoc = 0
	SET @isTransactionTypeAdd = 0
	SET @isTransactionTypeUpdate = 0
	SET @isTransactionTypeDelete = 0
	SET @isItem = 0
	SET @isBOM = 0
	SET @isUser = 0
	SET @isDocument = 0
	SET @isPurchaseDocument = 0
	SET @isSalesCreditNote = 0
    SET @isSalesInvoice = 0
	SET @IsPicklist = 0

	--Test object type (this can be found in the SDK 'BoObjectTypes Enumeration'
	IF (@object_type = N'20') BEGIN
		SET @isPurchaseDeliveryDocument = 1
	END ELSE IF (@object_type = N'PMX_MVHE') BEGIN
		SET @isMoveDoc = 1
	END ELSE IF (@object_type = N'4') BEGIN
		SET @isItem = 1
	END ELSE IF (@object_type = N'66') BEGIN
		SET @isBOM = 1
	END ELSE IF (@object_type = N'12') BEGIN
		SET @isUser = 1
	END ELSE IF (@object_type = N'15') BEGIN
		SET @isSalesDeliveryDocument = 1
	END ELSE IF (@object_type = N'59') BEGIN
		SET @isInventoryEntry = 1
	END ELSE IF (@object_type = N'60') BEGIN
		SET @isInventoryExit = 1
	END ELSE IF (@object_type = N'21') BEGIN
		SET @isPurchaseReturnDocument = 1
	END ELSE IF (@object_type = N'16') BEGIN
		SET @isSalesReturnDocument = 1
	END ELSE IF (@object_type = N'17') BEGIN
		SET @isSalesOrderDocument = 1
	END ELSE IF (@object_type = N'22') BEGIN
		SET @isPurchaseDocument = 1
	END ELSE IF (@object_type = N'14') BEGIN
		SET @isSalesCreditNote = 1
	END ELSE IF (@object_type = N'13') BEGIN
		SET @isSalesInvoice = 1
	END ELSE IF (@object_type = N'PMX_PLHE') BEGIN
		SET @IsPicklist = 1 
	END


	--Test transaction type
	IF (@transaction_type = N'A') BEGIN
		SET @isTransactionTypeAdd = 1
	END ELSE IF (@transaction_type = N'U') BEGIN
		SET @isTransactionTypeUpdate = 1
	END ELSE IF (@transaction_type = N'D') BEGIN
		SET @isTransactionTypeDelete = 1
	END

	--Set if it is a document
	IF ((@isPurchaseDeliveryDocument = 1) OR
		(@isSalesDeliveryDocument = 1) OR
		(@isInventoryEntry = 1) OR
		(@isInventoryExit = 1) OR
		(@isPurchaseReturnDocument = 1) OR
		(@isSalesReturnDocument = 1) OR
		(@isSalesOrderDocument = 1) OR
		(@isPurchaseDocument =1) OR
	    (@isSalesCreditNote = 1) OR
	    (@isSalesInvoice = 1) OR
		(@IsPicklist = 1)
		) BEGIN
		SET @isDocument = 1
	END

	--Check if we have to do something. If nothing to do then exit.	
	IF (NOT(
		(@isPurchaseDeliveryDocument=1 AND @isTransactionTypeAdd=1) OR	--Add of goods receipt(purchase delivery note) is supported
		(@isMoveDoc=1 AND @isTransactionTypeAdd=1) OR					--Add of move is supported
		(@isItem=1 AND (@isTransactionTypeAdd=1 OR @isTransactionTypeUpdate=1 OR @isTransactionTypeDelete=1))	OR													--All operations on an item are supported
		(@isBOM=1 AND (@isTransactionTypeAdd=1 OR @isTransactionTypeUpdate=1)) OR													--All operations on a BOM
		(@isUser=1 AND (@isTransactionTypeAdd=1 OR @isTransactionTypeUpdate=1))	OR													--All operations on a User
		--(@isSalesDeliveryDocument=1 AND @isTransactionTypeAdd=1) OR		--Add of delivery (sales delivery note) is supported
		(@isSalesDeliveryDocument=1) OR		--Add/Update of delivery (sales delivery note) is supported
		(@isInventoryEntry=1 AND @isTransactionTypeAdd=1) OR			--Add of inventory entry is supported
		(@isInventoryExit=1 AND @isTransactionTypeAdd=1) OR				--Add of inventory exit is supported
		(@isPurchaseReturnDocument=1 AND @isTransactionTypeAdd=1) OR	--Add of purchase return is supported
		(@isSalesReturnDocument=1 AND @isTransactionTypeAdd=1) OR		--Add of sales return is supported
		(@isSalesOrderDocument=1)OR										--Add/update of a sales order is supported
		(@isPurchaseDocument =1) OR		---Add/update of a purchase order is supported		
		(@isSalesCreditNote = 1) OR ---- Add/Update of a Sales Credit Note is supported
		(@isSalesInvoice = 1) OR ---Add/Update of a Sales Invoice is supported
		(@isPicklist = 1)		---Add/update of a picklist is supported 
		)) BEGIN
			RETURN
	END

	--The number of columns of the primary key must always be one. 
	IF @num_of_cols_in_key <> 1 BEGIN
		SET @error = 50001
		SET @error_message = 'Number of keys must be 1! Contact Produmex Support.'
		SELECT @error, @error_message
		EXEC xp_logevent @error, @error_message, ERROR
		RETURN
	END

	--General variabled
	DECLARE @keyAsInteger AS INT

	--**********************************************************************************************************
	--Do checks on an item
	--**********************************************************************************************************
	IF (@isItem=1) BEGIN
		
		--All checks are done so we can exit the checks.
		RETURN
	END

	--**********************************************************************************************************
	--Do checks on BOM
	--**********************************************************************************************************
	IF (@isBOM=1) BEGIN

		--All checks are done so we can exit the checks.
		RETURN
	END

	--**********************************************************************************************************
	--Do checks on a user
	--**********************************************************************************************************
	IF (@isUser=1) BEGIN

		--All checks are done so we can exit the checks.
		RETURN		
	END

	--Declare memory table for storing the document lines of a purchase delivery note, sales delivery note, ...
	DECLARE @documentLineTable TABLE (
		[DocEntry] [int] NOT NULL ,
		[LineNum] [int] NOT NULL ,
		[LineStatus] [nvarchar](1) NULL ,
		[ItemCode] [nvarchar](20) NULL ,
		[Dscription] [nvarchar](100) NULL,
		[Quantity] [numeric](19,6) NULL,
		[WhsCode] [nvarchar](8) NULL,
		[DocDate] [datetime] NULL,
		[U_PMX_LOCO] [nvarchar](max) NULL,
		[U_PMX_BBDT] [ntext] NULL,
		[U_PMX_LUID] [ntext] NULL,
		[U_PMX_QYSC] [nvarchar](max) NULL,
		[U_PMX_RIPI] [numeric](19,6) NULL,
		[U_PMX_ISLC] [nvarchar](max) NULL,
		[NumAtCard] [nvarchar] (100) NULL,
		[BatchNum]	[nvarchar](30) NULL
		
	)

	--**********************************************************************************************************
	--If the object that is changed is of the type document then
	--copy the document lines in a memory table
	--This memory table is used because it is easier and faster to do all the checks on this memory table
	--instead of getting the lines again from the database.
	--Checks are also done on the document lines
	--If a document line is indicated that it is a logistic carrier then this logistic carrier is put on the
	--logistic unit.
	--**********************************************************************************************************
	IF (@isDocument=1) BEGIN
		--Convert the list of key columns to an integer key
		SET @keyAsInteger = CONVERT(INT, @list_of_cols_val_tab_del)

		--**********************************************************************************************************
		--Copy document lines in the memory table
		--**********************************************************************************************************



	END
	--***********************************************************
	IF @isMoveDoc = 1
	BEGIN
		SET @keyAsInteger = CONVERT(INT, @list_of_cols_val_tab_del)


		-- attribute type not empty
		IF (exists (select * from 
				PMX_MVLI T0 WITH (NOLOCK)
			    JOIN [@PMX_TROY_BCQS] T1 WITH (NOLOCK)
					ON T0.SrcQualityStatusCode=T1.U_SrcQualityStatusCode and T0.DestQualityStatusCode=T1.U_DestQualityStatusCode and T1.U_CheckBatchAttribute='Y'
				JOIN PMX_ITBA T2 WITH (NOLOCK) 
					ON T2.ItriKey= T0.ItemTransactionalInfoKey and T1.U_Attribute = T2.BatchAttributeCode and T2.AttributeValue =''
				WHERE T0.DocEntry=@keyAsInteger
			))	
		BEGIN
				SET @error = 59000
				SET @error_message = 'Error on document: ' + @list_of_cols_val_tab_del + ' : batch attribute arrn''t filled in'
				SELECT @error, @error_message
				EXEC xp_logevent @error, @error_message, ERROR
				RETURN
		END

	END


	--**********************************************************************************************************
	--Do checks on a PurchaseOrder
	--**********************************************************************************************************
	

	--**********************************************************************************************************
	--Do checks on a SalesReturn 
	--**********************************************************************************************************





	--**********************************************************************************************************
	--Do checks on a SalesCreditNote
	--**********************************************************************************************************



	--**********************************************************************************************************
	--Do checks on a SalesInvoice
	--**********************************************************************************************************



	--**********************************************************************************************************
	--Do general checks
	--**********************************************************************************************************
	
	END


----Following code must be added to the stored procedure SBO_SP_TransactionNotification
----to execute the PMX stored procedure:

----**********************************************************************************************************
----Start executing Produmex Logex Addon code
----**********************************************************************************************************
--BEGIN TRY
--	EXEC [dbo].[PMX_SP_Troy_TransactionNotification] 
--	@object_type,
--	@transaction_type,
--	@num_of_cols_in_key,
--	@list_of_key_cols_tab_del,
--	@list_of_cols_val_tab_del,
--	@error = @error OUTPUT,
--	@error_message = @error_message OUTPUT
--END TRY
--BEGIN CATCH
--    SELECT @error = ERROR_NUMBER(), @error_message = ERROR_MESSAGE()
--    DECLARE @msg as NVARCHAR(255)
--    SET @msg = SUBSTRING('Error='+CAST(@error AS NVARCHAR) + ' Message=' + @error_message,1,255)
--    EXEC xp_logevent 999999, @msg  , ERROR
--END CATCH;
----**********************************************************************************************************
----End executing Produmex Logex Addon code
----**********************************************************************************************************















