using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Data;
using System.Data.Common;
using System.Globalization;
using System.Reflection;
using System.Text;
using Produmex.Foundation;
using Produmex.Foundation.Collections;
using Produmex.Foundation.Convertors;
using Produmex.Foundation.Data;
using Produmex.Foundation.Data.DataSetClient;
using Produmex.Foundation.Data.DbClient;
using Produmex.Foundation.Data.Sbo;
using Produmex.Foundation.Data.Sbo.BusinessObjects;
using Produmex.Foundation.Data.Sbo.BusinessObjects.Definitions.Tables;
using Produmex.Foundation.Data.Sbo.Providers;
using Produmex.Foundation.Data.Sbo.Utilities;
using Produmex.Foundation.Data.SqlClient;
using Produmex.Foundation.Diagnostics;
using Produmex.Foundation.Messages;
using Produmex.Foundation.SlimScreen;
using Produmex.Foundation.SlimScreen.Interfaces.Definitions;
using Produmex.Foundation.SlimScreen.Interfaces.Definitions.KnownDataSets;
using Produmex.Foundation.Utilities;
using Produmex.Foundation.Workflows;
using Produmex.Foundation.Workflows.Parameters;
using Produmex.Foundation.Wwf.Sbo.LocalServices;
using Produmex.Sbo.Logex.Data;
using Produmex.Sbo.Logex.Data.BusinessObjects;
using Produmex.Sbo.Logex.Data.BusinessObjects.Definitions;
using Produmex.Sbo.Logex.Data.BusinessObjects.Definitions.Tables;
using Produmex.Sbo.Logex.Data.DataObjects;
using Produmex.Sbo.Logex.Data.Providers;
using Produmex.Sbo.Logex.Data.Utilities;
using Produmex.Sbo.Logex.Troy.BusinessObjects.Definitions.Tables;


// PmxWorkflowExecutionTypes.SUB_FLOW
namespace Produmex.Sbo.Logex.WorkflowScripts
{
	/// <summary>
	/// Enter quantity
	/// </summary>
	public class WorkflowScript_WeighingTroyScript : WorkflowInstanceScriptBase
	{
		// Input parameters
		public ReadOnlyBinder<CultureInfo> DefaultCultureInfo;
		public ReadOnlyBinder<PmxOseCompany> PmxOseCompany;

		private static readonly ILog s_log = LogProvider.GetLogger(MethodInfo.GetCurrentMethod().DeclaringType);

		// Private fields
		private ISboProviderService m_sboProviderService;


		/// <summary>
		/// Definitions
		/// </summary>
		private static class Definitions
		{
			/// <summary>
			/// The stored procedure
			/// </summary>
			public const string STORED_PROCEDURE = "MCS_GetPotencyAdjustedBatches";

			/// <summary>
			///  The column name
			/// </summary>
			public const string COLUMN_NAME = "SAPNewPlannedQty";
		}

		#region Constructors

		public WorkflowScript_WeighingTroyScript(WorkflowInstanceBase parent, WorkflowInstanceFactory factory)
			: base(parent, factory)
		{
		}

		#endregion

		#region WorkflowInstanceScriptBase Members

		#endregion

		protected override void Execute()
		{
			// Parameters in scope
			Session session = GetScopeParameter("Session") as Session;
			m_sboProviderService = GetScopeParameter("<WwfService>ISboService") as ISboProviderService;

			WorkflowHelper methodHelper = new WorkflowScripts.WorkflowScript_WeighingTroyScript.WorkflowHelper
			{
				CultureInfo = this.DefaultCultureInfo.Get(),
				ShowScreen = session.ShowScreen,
				WaitForMessage = this.WaitForMessage,
				SboProviderService = m_sboProviderService
			};

			string deviceID = session.DeviceInstanceId;

			string weighingLocation = null;
			int? weighOrderKey = null;
			string produceItemCode = null;

			bool allowSwitchScale;
			bool allowOverruleWeigh;

			bool backToScan = false;

			PmxDictionary<int, double> uomCoefficients = new PmxDictionary<int, double>();
			WeighData weighData;

			PmxOseWarehouseInfo whsInfoDevice;
			StartWeigh(methodHelper, deviceID, out allowSwitchScale, out allowOverruleWeigh, out weighData, out whsInfoDevice);

			GotoResult result = GotoResult.SelectWeighingRoom;

			while (true)
			{
				GotoResult back;
				List<GotoResult> next;

				switch (result)
				{
					case GotoResult.SelectOrderToWeigh:
						back = GotoResult.SelectWeighingRoom;
						next = new List<GotoResult> { back, GotoResult.SelectWeighingRoom, GotoResult.SelectOrderToWeigh, GotoResult.SelectWeighOrderLine };
						result = GetOrderToWeigh(methodHelper, weighData.WeighingRoom.Code, out weighOrderKey, back);
						break;
					case GotoResult.SelectItemToWeigh:
						back = GotoResult.SelectWeighingRoom;
						next = new List<GotoResult> { back, GotoResult.SelectWeighingRoom, GotoResult.SelectWeighOrderLine };
						result = GetWeighableItem(methodHelper, weighData.WeighingRoom.Code, out produceItemCode, back);
						break;
					case GotoResult.CheckWeighStrategy:
						next = new List<GotoResult> { GotoResult.SelectItemToWeigh, GotoResult.SelectOrderToWeigh };
						result = CheckWeighStrategy(methodHelper, weighingLocation, weighData);
						produceItemCode = null;
						weighOrderKey = null;
						break;
					case GotoResult.SelectWeighingRoom:
						next = new List<GotoResult> { GotoResult.CheckWeighStrategy, GotoResult.Exit };
						result = GetWeighingRoom(methodHelper, whsInfoDevice.PmxWarehouseCode, deviceID, out weighingLocation);
						break;
					case GotoResult.SelectWeighOrderLine:
						back = produceItemCode == null ? GotoResult.SelectOrderToWeigh : GotoResult.SelectItemToWeigh;
						next = new List<GotoResult> { back, GotoResult.ScanSSCC, GotoResult.SelectWeighOrderLine };
						result = GetWeighOrderLineData(methodHelper, weighingLocation, produceItemCode, weighOrderKey, weighData, back);
						break;
					case GotoResult.WeighItem:
						backToScan = false;
						back = GotoResult.ScanSSCC;
						next = new List<GotoResult> { back, GotoResult.SelectScale, GotoResult.ScanSSCC, GotoResult.WeighItem, GotoResult.SelectOrderToWeigh, GotoResult.SelectItemToWeigh, GotoResult.SelectWeighOrderLine };
						result = WeighItem(session.DeviceInstanceId, methodHelper, weighData, uomCoefficients, allowOverruleWeigh, weighingLocation, produceItemCode, weighOrderKey, ref backToScan, back);
						break;
					case GotoResult.ScanSSCC:
						back = backToScan ? GotoResult.WeighItem : GotoResult.SelectWeighOrderLine;
						next = new List<GotoResult> { back, GotoResult.ScanSSCC, GotoResult.WeighItem };
						result = SelectSSCC(methodHelper, weighData, back);
						break;
					case GotoResult.SelectScale:
						next = new List<GotoResult> { GotoResult.SelectScale, GotoResult.WeighItem };
						result = SelectScale(methodHelper, weighingLocation, weighData, allowSwitchScale);
						break;
					case GotoResult.Exit:
					default:
						return;
				}

				if (!next.Contains(result))
					throw new Exception("Invalid next step:" + Enum.GetName(typeof(GotoResult), result));
			}
		}

		private void StartWeigh(WorkflowHelper methodHelper, string deviceID, out bool allowSwitchScale, out bool allowOverruleWeigh, out WeighData weighData, out PmxOseWarehouseInfo whsInfoDevice)
		{
			whsInfoDevice = m_sboProviderService.InvokeMethodWithDbConnection(false, false, null, null, GetPmxWarehouseInfo, deviceID);
			weighData = new WeighData();

			PmxDictionary<string, bool> authorizationsForLoggedInUser = GetAuthorizationsForUser(methodHelper, null, null, PmxUserAuthorizationColumnValues.WeighOrderOverruleSwitchScale.CODE, PmxUserAuthorizationColumnValues.WeighOrderOverruleWeight.CODE);

			authorizationsForLoggedInUser.TryGetValue(PmxUserAuthorizationColumnValues.WeighOrderOverruleSwitchScale.CODE, out allowSwitchScale);
			authorizationsForLoggedInUser.TryGetValue(PmxUserAuthorizationColumnValues.WeighOrderOverruleWeight.CODE, out allowOverruleWeigh);
		}

		private static GotoResult FinishProcess(WorkflowHelper methodHelper, string weighingLocation, int? weighOrderKey, string produceItemCode)
		{
			string query = BuildQuery.GetWeighOrderLines(weighingLocation, weighOrderKey, produceItemCode, methodHelper.SboProviderService.GetDbTool());
			DataSet dsLines = methodHelper.SboProviderService.RunView(false, null, null, query);

			if (GetNumberOfRows(dsLines) == 0)
			{
				if (produceItemCode == null)
				{
					return GotoResult.SelectOrderToWeigh;
				}
				else
				{
					return GotoResult.SelectItemToWeigh;
				}
			}
			else
			{
				return GotoResult.SelectWeighOrderLine;
			}
		}

		private static PmxDictionary<string, bool> GetAuthorizationsForUser(WorkflowHelper methodHelper, string userName, string password, params string[] authorizationCodes)
		{
			PmxDictionary<string, bool> authorizationsByCode = new PmxDictionary<string, bool>();

			methodHelper.SboProviderService.InvokeMethodWithDbConnection<object>(true, false, userName, password,
			   delegate (PmxDbConnection conn, object[] parameters)
			   {
				   PmxUserProvider usProv = new PmxUserProvider();
				   string userCode = SboHelper.SboUsername;
				   PmxDictionary<string, PmxUserGroupAuthorizationCodes> authorizations = usProv.GetAllPmxGroupAuthorizationsOfUser(userCode, null, conn);

				   foreach (string authorizationCode in authorizationCodes)
				   {
					   PmxUserGroupAuthorizationCodes userGroupAuthorizationCodes;

					   if (!authorizations.TryGetValue(authorizationCode, out userGroupAuthorizationCodes))
					   {
						   userGroupAuthorizationCodes = PmxUserGroupAuthorizationCodes.Disabled;
					   }

					   authorizationsByCode.Add(authorizationCode, userGroupAuthorizationCodes == PmxUserGroupAuthorizationCodes.Enabled);
				   }

				   return null;
			   });

			return authorizationsByCode;
		}

		private static bool? IsUserAuthorized(WorkflowHelper methodHelper, out string userName, out string password, string authorizationCode, PmxOverruleType overruleType)
		{
			GetCredentials(methodHelper, out userName, out password, overruleType);

			if (userName != null && password != null)
			{
				if (GetAuthorizationsForUser(methodHelper, userName, password, authorizationCode)[authorizationCode])
				{
					return true;
				}
			}
			else
			{
				//No user/paswd, so this means user wanted to go back
				return null;
			}

			return false;
		}

		private static bool? IsUserAuthorizedToSwitchScale(WorkflowHelper methodHelper, out string userName, out string password)
		{
			return IsUserAuthorized(methodHelper, out userName, out password, PmxUserAuthorizationColumnValues.WeighOrderOverruleSwitchScale.CODE, PmxOverruleType.SelectDifferentScale);
		}

		private static bool? IsUserAuthorizedToOverruleWeigh(WorkflowHelper methodHelper, out string userName, out string password, PmxOverruleType overruleType)
		{
			return IsUserAuthorized(methodHelper, out userName, out password, PmxUserAuthorizationColumnValues.WeighOrderOverruleWeight.CODE, overruleType);
		}

		private static void LogOverrule(WorkflowHelper methodHelper, string userName, string password, int weighOrderKey, int weighOrderLine, PmxOverruleType overruleType)
		{
			methodHelper.SboProviderService.InvokeMethodWithDbConnection<object>(true, false, userName, password,
							delegate (PmxDbConnection conn, object[] parameters)
							{
								PmxOverruleProvider overruleProvider = new PmxOverruleProvider(conn);
								PmxOverrule overruleBO = overruleProvider.GetNewBO();

								overruleBO.OverruleType = overruleType;
								overruleBO.BaseType = Data.BusinessObjects.Definitions.PmxObjectTypes.PmxWeighOrder;
								overruleBO.BaseEntry = weighOrderKey;
								overruleBO.BaseLine = weighOrderLine;
								overruleProvider.AddBO(overruleBO);
								return null;
							});
		}

		private static GotoResult SelectSSCC(WorkflowHelper methodHelper, WeighData weighData, GotoResult backGotoResult)
		{
			DataSet dsItems = new DataSet();
			DataSet dsPreparedItems = new DataSet();

			methodHelper.SboProviderService.InvokeMethodWithDbConnection<object>(false, false, null, null,
			delegate (PmxDbConnection dbConn, object[] parameters)
			{
				string query = BuildQuery.GetStocksForItem(weighData.WeighOrderKey, weighData.WeighingRoom.InputLocationCode, weighData.LineItemCode, weighData.WeighOrderLineKey, dbConn.DbTool);

				using (DbDataReader rdr = dbConn.CreateCommand(query).ExecuteReader())
				{
					DataTable tbl = dsItems.Tables.Add();
					tbl.Load(rdr);

					tbl = dsPreparedItems.Tables.Add();
					tbl.Load(rdr);
				}

				return null;
			});

			dsItems.Tables[0].Columns.Add(SelectItemToPrepareDataSet.Columns.TOTAL_PREPARED_QTY.Name, typeof(double));
			dsItems.Tables[0].Columns.Add(SelectItemToPrepareDataSet.Columns.INCOMPLETE_WEIGHING.Name, typeof(string));

			int i = -1;
			foreach (DataRow itemRow in dsItems.Tables[0].Rows)
			{
				string itemCode = Convert.ToString(itemRow[SelectItemToPrepareDataSet.Columns.ITEM_CODE.Name]);
				int lineNum = Convert.ToInt32(itemRow[SelectItemToPrepareDataSet.Columns.LINE_NUM.Name]);
				double qty = 0.0;

				foreach (DataRow preparedItemRow in dsPreparedItems.Tables[0].Rows)
				{
					if (itemCode.Equals(preparedItemRow[SelectItemToPrepareDataSet.Columns.ITEM_CODE.Name]) &&
						lineNum == Convert.ToInt32(preparedItemRow[SelectItemToPrepareDataSet.Columns.LINE_NUM.Name]))
					{
						qty += Convert.ToDouble(preparedItemRow[SelectItemToPrepareDataSet.Columns.QUANTITY.Name]);
					}
				}

				itemRow[SelectItemToPrepareDataSet.Columns.TOTAL_PREPARED_QTY.Name] = qty;
				itemRow[SelectItemToPrepareDataSet.Columns.INCOMPLETE_WEIGHING.Name] = false;
				if (!SystemInt32Convertor.ObjectToIntOrNull(itemRow[SelectItemToPrepareDataSet.Columns.LUID.Name]).HasValue)
				{
					//if there is multiple stock without luid, we need to find the selected one
					itemRow[SelectItemToPrepareDataSet.Columns.LUID.Name] = i;
					i--;
				}
			}

			methodHelper.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.Production.ISelectItemToPrepareScreen),
					methodHelper.CultureInfo, BuildParamCollection(
					"InitialErrorKey", methodHelper.InitialErrorKey,
					"TitleKey", "MSG_TITLE_SELECT_ITEM_TO_WEIGH",
					"DisableButton", "BTN_TIME",
					"AlwaysEnableNext", true,
					"ItemsDS", dsItems));

			methodHelper.Msg = methodHelper.WaitForMessage();
			methodHelper.InitialErrorKey = null;

			if (methodHelper.Msg.Name.EndsWith(".BackRequested"))
			{
				return backGotoResult;
			}

			if (methodHelper.Msg.Name.EndsWith(".RefreshRequested"))
			{
				return GotoResult.ScanSSCC;
			}

			if (!methodHelper.Msg.Name.EndsWith(".ItemSelected"))
			{
				methodHelper.InitialErrorKey = KnownMessageKeys.MSG_ERROR_OPERATION_NOT_SUPPORTED;

				return GotoResult.ScanSSCC;
			}

			// User selected an item

			string selectedItemCode = ExtractParameter<string>(methodHelper.Msg.Parameters, "itemCode");
			int? selectedItriKey = ExtractParameter<int?>(methodHelper.Msg.Parameters, "itriKey");
			int? selectedLUID = ExtractParameter<int?>(methodHelper.Msg.Parameters, "luid");
			int? productionOrderLineNum = ExtractParameter<int?>(methodHelper.Msg.Parameters, "lineNum");

			//int? luid = ExtractParameter<int?>(methodHelper.Msg.Parameters, "luid");
			//int? lineNumber = ExtractParameter<int?>(methodHelper.Msg.Parameters, "lineNum");

			//string filter = SelectItemToPrepareDataSet.Columns.LUID.Name + " = " + luid;

			DataRow[] selectedItemRows = dsItems.Tables[0].Select(string.Format("{0} = {1} AND {2} {3} {4}  AND {5} {6} {7} AND {8} = {9}",
				 DataSetCommandHelper.EscapeColumnName(SelectItemToPrepareDataSet.Columns.ITEM_CODE.Name),
				 DataSetCommandHelper.FormatAndEscapeColumnValue(selectedItemCode, methodHelper.CultureInfo),
				 DataSetCommandHelper.EscapeColumnName(SelectItemToPrepareDataSet.Columns.ITRI_KEY.Name),
				 selectedItriKey.HasValue ? "=" : "",
				 selectedItriKey.HasValue ? DataSetCommandHelper.FormatAndEscapeColumnValue(selectedItriKey.Value, methodHelper.CultureInfo) : "IS NULL",
				 DataSetCommandHelper.EscapeColumnName(SelectItemToPrepareDataSet.Columns.LUID.Name),
				 selectedLUID.HasValue ? "=" : "",
				 selectedLUID.HasValue ? DataSetCommandHelper.FormatAndEscapeColumnValue(selectedLUID.Value, methodHelper.CultureInfo) : "IS NULL",
				 DataSetCommandHelper.EscapeColumnName(SelectItemToPrepareDataSet.Columns.LINE_NUM.Name),
				 productionOrderLineNum ?? -1
				));


			DataSet ds = dsItems.Clone();
			foreach (DataRow dr in selectedItemRows)
			{
				ds.Tables[0].Rows.Add(dr.ItemArray);
			}

			DataRow row = ds.Tables[0].Rows[0];

			if (SystemInt32Convertor.ObjectToInt(row[SelectItemToPrepareDataSet.Columns.LUID.Name], -1, -1) < 0)
			{
				row[SelectItemToPrepareDataSet.Columns.LUID.Name] = DBNull.Value;
			}

			SelectSSCC_FillWeighData(weighData, row);

			if (weighData.QtyAvailable > 0)
			{
				return GotoResult.WeighItem;
			}
			else
			{
				methodHelper.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
					methodHelper.CultureInfo,
					BuildParamCollection(
						"MessageKey", "MSG_ERROR_ITEM_HAS_NO_STOCK",
						"IsError", true,
						"Parameters", new object[] {weighData.LineItemDescription},
						"ShowButton", true));
				methodHelper.WaitForMessage();
				return GotoResult.ScanSSCC;
			}
		}

		private static void SelectSSCC_FillWeighData(WeighData weighData, DataRow row)
		{
			weighData.BestBeforeDate = SystemDateTimeConvertor.ObjectToDateTimeOrNull(row[SelectItemToPrepareDataSet.Columns.BEST_BEFORE_DATE.Name], false);
			weighData.BatchNumber = Convert.ToString(row[SelectItemToPrepareDataSet.Columns.BATCH_NUMBER.Name]);
			weighData.BatchNumber2 = Convert.ToString(row[SelectItemToPrepareDataSet.Columns.BATCH_NUMBER2.Name]);
			weighData.StockSSCC = SystemStringConvertor.ObjectToStringOrNull(row[SelectItemToPrepareDataSet.Columns.SSCC.Name]);
			weighData.StockLuid = SystemInt32Convertor.ObjectToIntOrNull(row[SelectItemToPrepareDataSet.Columns.LUID.Name]);
			weighData.QualityStatus = Convert.ToString(row[SelectItemToPrepareDataSet.Columns.QUALITY_STATUS.Name]);
			weighData.ItriKey = SystemInt32Convertor.ObjectToIntOrNull(row[SelectItemToPrepareDataSet.Columns.ITRI_KEY.Name]);

			double qtyTodo = Convert.ToDouble(row[SelectItemToPrepareDataSet.Columns.QTY_PLANNED.Name]) - Convert.ToDouble(row[SelectItemToPrepareDataSet.Columns.QTY_ISSUED.Name]);
			weighData.QtyToDo = qtyTodo > 0 ? qtyTodo : 0;
			double? qtyAvailable = SystemDoubleConvertor.ObjectToDoubleOrNull(row[SelectItemToPrepareDataSet.Columns.QUANTITY.Name]);
			weighData.QtyAvailable = qtyAvailable ?? 0.0;
			weighData.QtyPlanned = Convert.ToDouble(row[SelectItemToPrepareDataSet.Columns.QTY_PLANNED.Name]);
			weighData.Tolerance = SystemDoubleConvertor.ObjectToDouble(row[SelectItemToPrepareDataSet.Columns.TOLERANCE_PCT.Name], 0.0, 0.0);

			weighData.InventoryUom = Convert.ToString(row[SelectItemToPrepareDataSet.Columns.UOM.Name]);
			weighData.InventoryUomDecimals = Convert.ToInt32(row[SelectItemToPrepareDataSet.Columns.UOM_DECIMALS.Name]);
		}

		private static GotoResult SelectScale(WorkflowHelper methodHelper, string weighingLocation, WeighData weighData, bool allowScaleSwitch)
		{
			string query = SboCommands.GetQueryGetAllScalesOfWeighingRooms(weighingLocation, true, methodHelper.CultureInfo);

			DataSet dsScales = methodHelper.SboProviderService.RunView(false, null, null, query);

			weighData.NumberOfScales = dsScales.Tables[0].Rows.Count;

			methodHelper.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.ISelectCodeNameScreen),
			   methodHelper.CultureInfo, BuildParamCollection(
				"InitialErrorKey", methodHelper.InitialErrorKey,
				"TitleKey", "MSG_TITLE_SELECT_SCALE",
				"CodeNameDS", dsScales
				));

			methodHelper.Msg = methodHelper.WaitForMessage();

			methodHelper.InitialErrorKey = null;

			if (methodHelper.Msg.Name.EndsWith(".CodeNameSelected"))
			{
				string scaleCode = ExtractParameter<string>(methodHelper.Msg.Parameters, "code");

				if (scaleCode != weighData.ScaleCode)
				{
					string userName = null;
					string password = null;

					bool? allowOperation = allowScaleSwitch ? true : IsUserAuthorizedToSwitchScale(methodHelper, out userName, out password);

					if (allowOperation ?? false)
					{
						LogOverrule(methodHelper, userName, password, weighData.WeighOrderKey, weighData.WeighOrderLineNum, PmxOverruleType.SelectDifferentScale);

						weighData.ScaleCode = scaleCode;
					}
					else
					{
						if (allowOperation.HasValue)
						{
							methodHelper.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
								   methodHelper.CultureInfo,
								   BuildParamCollection(
									   "MessageKey", "MSG_ERROR_USER_IS_NOT_AUTHORIZED_TO_SWITCH_SCALES",
									   "ShowButton", true));
							methodHelper.WaitForMessage();
						}
						return GotoResult.SelectScale;
					}
				}

				return GotoResult.WeighItem;
			}

			return GotoResult.WeighItem;
		}

		private static void GetCredentials(WorkflowHelper methodHelper, out string userName, out string userPassword, PmxOverruleType overruleType)
		{
			userName = null;
			userPassword = null;

			string titleKey = "MSG_TITLE_ENTER_CREDENTIALS_OVERRULE_" + overruleType.ToString().ToUpper();

			methodHelper.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.ILogonScreen),
				methodHelper.CultureInfo, BuildParamCollection(
					"DefaultUserName", "",
					"TitleKey", titleKey,
					"IsForCredentialCheck", true));
			methodHelper.Msg = methodHelper.WaitForMessage();

			if (methodHelper.Msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.ILogonScreenReply.CredentialsEntered")
			{
				userName = ExtractParameter<string>(methodHelper.Msg.Parameters, "UserName");
				userPassword = ExtractParameter<string>(methodHelper.Msg.Parameters, "Password");

				try
				{
					methodHelper.SboProviderService.InvokeMethodWithDbConnection<object>(true, false, userName, userPassword,
					 delegate
					 {
						 return null;
					 });
				}
				catch (Exception ex)
				{
					ShowMessageScreenForException(methodHelper, "PmxError", ((ProdumexException)ex).Error, "ShowButton", true);
					userName = null;
					userPassword = null;
				}
			}
		}

		private static void ShowMessageScreenForException(WorkflowHelper methodHelper, string name1, object value1, params object[] parameters)
		{
			methodHelper.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
			methodHelper.CultureInfo,
			BuildParamCollection(name1, value1, parameters));

			methodHelper.Msg = methodHelper.WaitForMessage();
		}

		private static GotoResult WeighItem(string deviceInstanceId, WorkflowHelper methodHelper, WeighData data, PmxDictionary<int, double> uomCoefficients, bool allowWeighOverrule, string weighingLocation, string produceItemCode, int? weighOrderKey, ref bool backToScan, GotoResult backGotoResult)
		{
			if (data.ScaleData == null || data.ScaleCode != data.ScaleData.Code)
			{
				WeighItem_GetScaleData(methodHelper, data, uomCoefficients);
			}

			WeighItem_GetSSCCData(methodHelper, data);

			double[] weighingTarget = {
				Math.Max( 0.0,  data.QtyToDo ),
				data.QtyPlanned*(data.Tolerance / 100.0) };

			string tareCommand = null;
			string zeroCommand = null;
			string weighCommand = null;
			string weighResultFormat = null;

			if (data.ScaleData != null)
			{
				foreach (PmxOseScaleCommand scaleCommand in data.ScaleData.ScaleCommands)
				{
					switch (scaleCommand.CommandType)
					{
						case PmxScaleCommandType.GetWeight:
							weighCommand = scaleCommand.CommandFormat;
							break;
						case PmxScaleCommandType.SetTare:
							tareCommand = scaleCommand.CommandFormat;
							break;
						case PmxScaleCommandType.SetZero:
							zeroCommand = scaleCommand.CommandFormat;
							break;
						case PmxScaleCommandType.WeightReturn:
							weighResultFormat = scaleCommand.CommandFormat;
							break;
						case PmxScaleCommandType.TareReturn:
							break;
						case PmxScaleCommandType.ZeroReturn:
							break;
						default:
							break;
					}
				}
			}

			methodHelper.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.Weighing.IWeighCaptureWeighOrderScreen),
				methodHelper.CultureInfo, BuildParamCollection(
				"InitialErrorKey", methodHelper.InitialErrorKey,
				"TitleKey", "MSG_TITLE_WEIGH_ITEM",
				"ProdOrderDocNum", data.ProductionOrderDocNum,
				"ProductionItemCode", data.ProductionOrderItemCode,
				"ProductionItemDescr", data.ProductionOrderItemDescription,
				"OrderLineItemCode", data.LineItemCode,
				"OrderLineItemDescr", data.LineItemDescription,
				"WeighOrderNum", data.WeighOrderKey,
				"WeighOrderLineNum", data.WeighOrderLineKey,
				"BatchNum", data.BatchNumber,
				"BatchNum2", data.BatchNumber2,
				"BestBeforeDate", string.Format("{0:d}", data.BestBeforeDate),
				"StockSSCC", data.StockSSCC,
				"Uom", data.ScaleUom ?? data.InventoryUom,
				"UomDecimals", data.UomDecimals ?? data.InventoryUomDecimals,
				"UomCoefficient", data.ScaleCoefficient,
				"QtyPlanned", data.QtyPlanned,
				"QtyAvailable", data.QtyAvailable,
				"QtyToDo", data.QtyToDo,
				"TargetLuidDS", data.TargetLuidDS,
				"SelectedLuid", data.SelectedLuid,
				"WeighingTarget", weighingTarget,
				"NumberOfScales", data.NumberOfScales,
				"ScaleName", data.ScaleData?.Name,
				"ScaleProviderType", data.ScaleData != null ? data.ScaleData.UseAsService ? "Service" : "GenericScale" : null,
				"TareCommand", tareCommand,
				"ZeroCommand", zeroCommand,
				"WeighCommand", weighCommand,
				"WeighResultFormat", weighResultFormat,
				"ScaleSettings", data.ScaleData?.ScaleDefinition
				));

			methodHelper.Msg = methodHelper.WaitForMessage();
			methodHelper.InitialErrorKey = null;

			if (methodHelper.Msg.Name.EndsWith(".BackRequested"))
			{
				return backGotoResult;
			}
			else if (methodHelper.Msg.Name.EndsWith(".SelectScale"))
			{
				int? targetLuid = ExtractParameter<int?>(methodHelper.Msg.Parameters, "targetLuid");

				data.SelectedLuid = targetLuid;

				return GotoResult.SelectScale;
			}
			else if (methodHelper.Msg.Name.EndsWith(".SelectStock"))
			{
				double weight = ExtractParameter<double>(methodHelper.Msg.Parameters, "weight");
				weight = Produmex.Foundation.Utilities.FormattingHelper.RoundSAPDecimal(weight, data.InventoryUomDecimals ?? 0);
				bool removeRemainingStock = ExtractParameter<bool>(methodHelper.Msg.Parameters, "removeRemainingStock");
				int? targetLuid = ExtractParameter<int?>(methodHelper.Msg.Parameters, "targetLuid");

				data.SelectedLuid = targetLuid;

				if (weight > 0.0)
				{
					ProcessSelectStock(methodHelper, data, weight, removeRemainingStock, targetLuid);
				}

				backToScan = true;
				return GotoResult.ScanSSCC;
			}
			else if (methodHelper.Msg.Name.EndsWith(".Weigh"))
			{
				double weight = ExtractParameter<double>(methodHelper.Msg.Parameters, "weight");
				weight = Produmex.Foundation.Utilities.FormattingHelper.RoundSAPDecimal(weight, data.InventoryUomDecimals ?? 0);
				int? targetLuid = ExtractParameter<int?>(methodHelper.Msg.Parameters, "targetLuid");

				data.SelectedLuid = targetLuid;

				ProcessWeigh(methodHelper, data, weight, targetLuid, deviceInstanceId);

				return GotoResult.WeighItem;
			}
			else if (methodHelper.Msg.Name.EndsWith(".Finish"))
			{
				int? targetLuid = ExtractParameter<int?>(methodHelper.Msg.Parameters, "targetLuid");

				data.SelectedLuid = targetLuid;

				if (ProcessFinish(methodHelper, data, allowWeighOverrule))
				{
					return FinishProcess(methodHelper, weighingLocation, weighOrderKey, produceItemCode);
				}
				else
				{
					return GotoResult.WeighItem;
				}
			}

			return backGotoResult;
		}

		private static void WeighItem_GetScaleData(WorkflowHelper methodHelper, WeighData data, PmxDictionary<int, double> uomCoefficients)
		{
			data.ScaleData = methodHelper.SboProviderService.InvokeMethodWithDbConnection(true, false, null, null,
						delegate (PmxDbConnection conn, object[] parameters)
						{
							SboUomProvider prov = new SboUomProvider((SboConnection)conn);
							int uomEntry = prov.GetUomEntryOfUomCode(data.InventoryUom);

							PmxOseScaleProvider scaleProvider = new PmxOseScaleProvider(conn);

							PmxOseScale scale = scaleProvider.GetScaleForLocationWithUomAndQuantity(data.ScaleCode, data.WeighingRoom.Code, uomCoefficients, uomEntry, data.QtyPlanned);

							if (scale != null)
							{
								data.ScaleCoefficient = prov.GetCoefficientForUom(scale.UomEntry, uomEntry, uomCoefficients);
								data.ScaleUom = prov.GetUomCodeOfUomEntry(scale.UomEntry);
								if (!data.ScaleCoefficient.HasValue)
								{
									s_log.Error(string.Format("There is no conversion between {0} and {1}", data.InventoryUom, data.ScaleUom));
								}

								NumberFormatInfo nfi = new NumberFormatInfo();
								nfi.NumberDecimalSeparator = ".";

								string[] accuracyComponents = scale.Accuracy.ToString(nfi).Split('.');

								data.UomDecimals = accuracyComponents.Length == 2 ? accuracyComponents[1].Length : 0;
							}
							return scale;
						});
		}

		private static void WeighItem_GetSSCCData(WorkflowHelper methodHelper, WeighData data)
		{
			DataSet dsPreparedItems = new DataSet();

			methodHelper.SboProviderService.InvokeMethodWithDbConnection<object>(false, false, null, null,
			delegate (PmxDbConnection dbConn, object[] parameters)
			{
				string query = BuildQuery.GetPreparedQuantitites(data.WeighOrderKey, data.WeighOrderLineKey, dbConn.DbTool);
				using (DbDataReader rdr = dbConn.CreateCommand(query).ExecuteReader())
				{
					DataTable tbl = dsPreparedItems.Tables.Add();
					tbl.Load(rdr);
				}
				return null;
			});

			DataSet dsTargetLuid = new DataSet();
			DataTable tblTargetLuid = dsTargetLuid.Tables.Add();
			tblTargetLuid.Columns.Add("InternalKey", typeof(int));
			tblTargetLuid.Columns.Add("SSCC", typeof(string));
			tblTargetLuid.Columns.Add("Quantity", typeof(double));
			tblTargetLuid.Columns.Add("Coeficient", typeof(double));
			tblTargetLuid.Columns.Add("UomDecimals", typeof(int));
			tblTargetLuid.Columns.Add("Uom", typeof(string));
			tblTargetLuid.Rows.Add(null, "<new>");

			PmxDictionary<int, DataRow> targetLuids = new PmxDictionary<int, DataRow>();

			foreach (DataRow row in dsPreparedItems.Tables[0].Select(string.Format("{0} = {1} AND {2} = {3} AND {4} = {5}",
					DataSetCommandHelper.EscapeColumnName(SelectItemToPrepareDataSet.Columns.ITEM_CODE.Name),
					DataSetCommandHelper.FormatAndEscapeColumnValue(data.LineItemCode, methodHelper.CultureInfo),
					DataSetCommandHelper.EscapeColumnName(SelectItemToPrepareDataSet.Columns.STORAGE_LOCATION.Name),
					DataSetCommandHelper.FormatAndEscapeColumnValue(data.WeighingRoom.Code, methodHelper.CultureInfo),
					DataSetCommandHelper.EscapeColumnName(SelectItemToPrepareDataSet.Columns.LINE_NUM.Name),
					DataSetCommandHelper.FormatAndEscapeColumnValue(data.ProductionOrderLineNum, methodHelper.CultureInfo))))
			{
				if (!Tester.IsNullOrEmpty(row[SelectItemToPrepareDataSet.Columns.LUID.Name]))
				{
					int luid = Convert.ToInt32(row[SelectItemToPrepareDataSet.Columns.LUID.Name]);

					DataRow targetLuidRow;

					if (!targetLuids.TryGetValue(luid, out targetLuidRow))
					{
						targetLuids.Add(luid, tblTargetLuid.Rows.Add(
							luid,
							row[SelectItemToPrepareDataSet.Columns.SSCC.Name],
							row[SelectItemToPrepareDataSet.Columns.QUANTITY.Name],
							data.ScaleCoefficient ?? 1.0,
							data.UomDecimals,
							data.ScaleUom));
					}
					else
					{
						double qty = Convert.ToDouble(row[SelectItemToPrepareDataSet.Columns.QUANTITY.Name]);
						targetLuidRow["Quantity"] = Convert.ToDouble(targetLuidRow["Quantity"]) + qty;
					}
				}
			}

			data.TargetLuidDS = dsTargetLuid;
		}

		private static bool ProcessFinish(WorkflowHelper methodHelper, WeighData data, bool allowWeighOverrule)
		{
			try
			{
				double totalQuantity = 0.0;

				List<int> luids = new List<int>();

				foreach (DataRow itemRow in data.TargetLuidDS.Tables[0].Rows)
				{
					int? luid = SystemInt32Convertor.ObjectToIntOrNull(itemRow["InternalKey"]);
					if (luid.HasValue && luid.Value != 0)
					{
						totalQuantity += Convert.ToDouble(itemRow[SelectItemToPrepareDataSet.Columns.QUANTITY.Name]);
						luids.Add(luid.Value);
					}
				}

				bool needOverruleAuthorization = false;
				PmxOverruleType overruleType = PmxOverruleType.WeighLess;

				if (data.QtyPlanned * (1 - data.Tolerance / 100) > totalQuantity)
				{
					needOverruleAuthorization = true;
					overruleType = PmxOverruleType.WeighLess;
				}
				else if (data.QtyPlanned * (1 + data.Tolerance / 100) < totalQuantity)
				{
					needOverruleAuthorization = true;
					overruleType = PmxOverruleType.WeighMore;
				}

				if (needOverruleAuthorization)
				{
					string userName = null;
					string password = null;

					bool? allowOperation = allowWeighOverrule ? true : IsUserAuthorizedToOverruleWeigh(methodHelper, out userName, out password, overruleType);

					if (allowOperation ?? false)
					{
						LogOverrule(methodHelper, userName, password, data.WeighOrderKey, data.WeighOrderLineNum, overruleType);

						CloseWeighOrderAndMoveLuids(methodHelper, data, luids);
					}
					else
					{
						if (allowOperation.HasValue)
						{
							methodHelper.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
								   methodHelper.CultureInfo,
								   BuildParamCollection(
									   "MessageKey", "MSG_ERROR_USER_IS_NOT_AUTHORIZED_TO_WEIHG_OUTSIDE_THE_TOLERANCE_RANGE",
									   "ShowButton", true));
							methodHelper.WaitForMessage();
						}
						return false;
					}
				}
				else
				{
					CloseWeighOrderAndMoveLuids(methodHelper, data, luids);
				}
			}
			catch (ProdumexException ex)
			{
				s_log.Error(ex.ToString());

				ShowMessageScreenForException(methodHelper, "PmxError", ex.Error, "ShowButton", true);
				methodHelper.InitialErrorKey = null;

				return false;
			}
			catch (Exception ex)
			{
				s_log.Error(ex.ToString());

				ShowMessageScreenForException(methodHelper, "MessageKey", ExceptionHelper.Format(ex, methodHelper.CultureInfo),
						"NoTranslationOfMessageKey", true,
						"IsError", true,
						"ShowButton", true);
				methodHelper.InitialErrorKey = null;

				return false;
			}

			return true;
		}

		private static void CloseWeighOrderAndMoveLuids(WorkflowHelper methodHelper, WeighData data, List<int> luids)
		{
			methodHelper.SboProviderService.InvokeMethodWithDbConnection<object>(false, false, null, null,
					delegate (PmxDbConnection dbConn, object[] parameters)
					{
						PmxWeighOrderProvider woProvider = new PmxWeighOrderProvider(dbConn);
						woProvider.CloseWeighOrderLine(data.WeighOrderKey, data.WeighOrderLineNum);

						PmxMoveProvider prov = new PmxMoveProvider(dbConn);
						prov.MakeAdHocMoveForLUIDs(luids, data.WeighingRoom.OutputLocationCode, Data.BusinessObjects.Definitions.PmxObjectTypes.PmxWeighOrder, data.WeighOrderKey);
						return null;
					});
		}

		private static void ProcessSelectStock(WorkflowHelper methodHelper, WeighData data, double weight, bool removeRemainingStock, int? targetLuid)
		{
			try
			{
				bool needsReceipt = weight > data.QtyAvailable;

				methodHelper.SboProviderService.InvokeMethodWithDbConnection<object>(needsReceipt || removeRemainingStock, false, null, null,  // Use SBO when theoQtyCorrection is needed
						   delegate (PmxDbConnection dbConn, object[] parameters)
						   {
							   int dstLuid = MoveStockFromInputLocationToWeightRoom(data, weight, targetLuid, needsReceipt, dbConn);

							   if (removeRemainingStock)
							   {
								   ProcessSelectStock_RemoveRemainingStock(data, weight, dbConn);
							   }

							   data.SelectedLuid = dstLuid;

							   return null;
						   });
			}
			catch (ProdumexException ex)
			{
				s_log.Error(ex.ToString());

				ShowMessageScreenForException(methodHelper, "PmxError", ex.Error, "ShowButton", true);
				methodHelper.InitialErrorKey = null;
			}
			catch (Exception ex)
			{
				s_log.Error(ex.ToString());

				ShowMessageScreenForException(methodHelper, "MessageKey", ExceptionHelper.Format(ex, methodHelper.CultureInfo),
						"NoTranslationOfMessageKey", true,
						"IsError", true,
						"ShowButton", true);
				methodHelper.InitialErrorKey = null;
			}
		}

		private static void ProcessSelectStock_RemoveRemainingStock(WeighData data, double weight, PmxDbConnection dbConn)
		{
			PmxProductionIssueProvider prdProv = new PmxProductionIssueProvider((SboConnection)dbConn);
			PmxOseProvider oseProv = new PmxOseProvider(dbConn);
			//add a stock issue
			using (PmxProductionIssue prdIssue = prdProv.GetNewBO())
			{
				DocumentLineInfo line = new DocumentLineInfo();

				ProcessSelectStock_FillDocumentLineInfoForProductionIssueLine(data, weight, oseProv, line);

				DocumentInfoHelper.AddDocumentLineInfoToBO(line, prdIssue, prdProv);

				if (prdIssue.Lines.Count > 0 &&
					prdIssue.Lines[0].Quantity > 0)
				{
					prdProv.AddBO(prdIssue);
				}
			}
		}

		private static void ProcessSelectStock_FillDocumentLineInfoForProductionIssueLine(WeighData data, double weight, PmxOseProvider oseProv, DocumentLineInfo line)
		{
			line.StorageLocationCode = data.WeighingRoom.Code;

			line.PmxBaseEntry = data.WeighOrderKey;
			line.PmxBaseLine = data.WeighOrderLineNum;
			line.PmxBaseType = Data.BusinessObjects.Definitions.PmxObjectTypes.PmxWeighOrder;

			line.BatchNumber1 = data.BatchNumber;
			line.BatchNumber2 = data.BatchNumber2;
			line.BestBeforeDate = data.BestBeforeDate;

			line.ItemCode = data.LineItemCode;

			line.LogisticUnitIdentKey = data.StockLuid;

			line.SSCC = data.StockSSCC;
			line.QualityStatusCode = data.QualityStatus;
			line.Quantity = data.QtyAvailable - weight;
			line.QuantityPerUom = 1; // Always one for stock
			line.StorageLocationCode = data.WeighingRoom.InputLocationCode;

			PmxOseWarehouseInfo whsInfo = oseProv.GetPmxWarehouseInfo(line.StorageLocationCode);
			line.WarehouseCode = whsInfo.SboWarehouseCode;
		}

		private static int MoveStockFromInputLocationToWeightRoom(WeighData data, double weight, int? targetLuid, bool needsReceipt, PmxDbConnection dbConn)
		{
			int dstLuid;

			// Determine dstLuid
			if (targetLuid.HasValue)
			{
				dstLuid = targetLuid.Value;
			}
			else
			{
				PmxLogisticUnitIDProvider luidProv = new PmxLogisticUnitIDProvider(dbConn);
				dstLuid = luidProv.GenerateNewLogisticUnit(null);
			}

			// Execute move of available qty
			PmxMoveProvider moveProv = new PmxMoveProvider(dbConn);

			List<MoveItemInfo> infos = new List<MoveItemInfo>();

			MoveItemInfo info = new MoveItemInfo();
			MoveStockFromInputLocationToWeightRoom_FillMoveItemInfo(data, weight, dstLuid, info);

			infos.Add(info);

			moveProv.MakeMoveFromMoveItemInfoList(infos);

			if (needsReceipt)
			{
				PmxProductionReceiptProvider receiptProv = new PmxProductionReceiptProvider((SboConnection)dbConn);
				using (PmxProductionReceipt receipt = receiptProv.GetNewBO())
				{
					DocumentLineInfo receiptLine = new DocumentLineInfo();
					double quantityToAdd = weight - data.QtyAvailable;
					MoveStockFromInputLocationToWeightRoom_FillReceiptLine(data, dbConn, info, receiptLine, quantityToAdd);

					PmxProductionReceiptLine sboReceiptLine = DocumentInfoHelper.AddDocumentLineInfoToBO(receiptLine, receipt, receiptProv);

					//we need to reduce the open quantity for the weigh order for the weight that was added
					PmxWeighOrderProvider woProv = new PmxWeighOrderProvider(dbConn);
					woProv.ConsumeWeighOrderLineByReceiptLine(sboReceiptLine);

					receiptProv.AddBO(receipt);
				}
			}

			if (!targetLuid.HasValue)
			{
				MoveStockFromInputLocationToWeightRoom_AddProductionOrderPreparedComponent(data, weight, dbConn, dstLuid);
			}
			else
			{
				CollectionWithDispose<PmxProductionOrderPreparedComponent> comps = new CollectionWithDispose<PmxProductionOrderPreparedComponent>();
				string whereClause = BuildQuery.GetProductionOrderPreparedComponentWhereClause(data.ProductionOrderDocEntry, null, dstLuid, data.LineItemCode, data.ProductionOrderLineNum, dbConn);
				PmxInternalBusinessObjectHelper.GetBOsByWhereClause(s_log, dbConn, typeof(PmxProductionOrderPreparedComponent), whereClause, null, true, comps);

				if (comps.Count > 0)
				{
					PmxProductionOrderPreparedComponentProvider pcProv = new PmxProductionOrderPreparedComponentProvider(dbConn);
					comps[0].Quantity += weight;
					pcProv.UpdateBO(comps[0], false, false);
				}
				else
				{
					MoveStockFromInputLocationToWeightRoom_AddProductionOrderPreparedComponent(data, weight, dbConn, dstLuid);
				}
			}

			return dstLuid;
		}

		private static void MoveStockFromInputLocationToWeightRoom_AddProductionOrderPreparedComponent(WeighData data, double weight, PmxDbConnection dbConn, int dstLuid)
		{
			PmxProductionOrderPreparedComponentProvider pcProv = new PmxProductionOrderPreparedComponentProvider(dbConn);
			using (PmxProductionOrderPreparedComponent newPc = pcProv.GetNewBO())
			{
				newPc.ProdOrderDocEntry = data.ProductionOrderDocEntry;
				newPc.ProdBatchSeqNr = null;
				newPc.LogUnitIdentKey = dstLuid;
				newPc.ItemCode = data.LineItemCode;
				newPc.Quantity = weight;
				newPc.ProdOrderLineNum = data.ProductionOrderLineNum;

				pcProv.AddBO(newPc);
			}
		}

		private static void MoveStockFromInputLocationToWeightRoom_FillReceiptLine(WeighData data, PmxDbConnection dbConn, MoveItemInfo info, DocumentLineInfo receiptLine, double quantityToAdd)
		{
			receiptLine.ItemCode = info.ItemCode;
			receiptLine.ItemTransactionalInfoKey = info.ItemTransactionalInfoKey;

			PmxItemTransactionalInfoProvider itriProv = new PmxItemTransactionalInfoProvider(dbConn);

			if (info.ItemTransactionalInfoKey.HasValue)
			{
				PmxItemTransactionalInfo itriInfo = itriProv.GetBO(info.ItemTransactionalInfoKey.Value);
				receiptLine.BatchNumber1 = itriInfo.BatchNumber1;
				receiptLine.BatchNumber2 = itriInfo.BatchNumber2;
				receiptLine.BestBeforeDate = itriInfo.BestBeforeDate;
			}

			receiptLine.QuantityPerUom = 1;
			receiptLine.Quantity = quantityToAdd;
			receiptLine.Uom = data.ScaleUom;
			receiptLine.WarehouseCode = data.WhsInfo.SboWarehouseCode;
			receiptLine.LogisticUnitIdentKey = info.DestinationLogisticUnitIdentKey;
			receiptLine.QualityStatusCode = info.DestinationQualityStatusCode;
			receiptLine.StorageLocationCode = info.DestinationLocationCode;
			receiptLine.PmxBaseEntry = data.WeighOrderKey;
			receiptLine.PmxBaseLine = data.WeighOrderLineNum;
			receiptLine.PmxBaseType = Data.BusinessObjects.Definitions.PmxObjectTypes.PmxWeighOrder;
		}

		private static void MoveStockFromInputLocationToWeightRoom_FillMoveItemInfo(WeighData data, double weight, int dstLuid, MoveItemInfo info)
		{
			info.BaseType = Data.BusinessObjects.Definitions.PmxObjectTypes.PmxWeighOrder;
			info.BaseEntry = data.WeighOrderKey;
			info.BaseLine = data.WeighOrderLineNum;
			info.ItemCode = data.LineItemCode;
			info.ItemDescription = data.LineItemName;
			info.ItemTransactionalInfoKey = data.ItriKey;
			info.Quantity = weight > data.QtyAvailable ? data.QtyAvailable : weight;
			info.QuantityPerUom = 1;
			info.Uom = data.InventoryUom;
			info.SourceLogisticUnitIdentKey = data.StockLuid;
			info.SourceLocationCode = data.WeighingRoom.InputLocationCode;
			info.SourcePmxWhsCode = data.WeighingRoom.PmxWarehouseCode;
			info.SourceQualityStatusCode = data.QualityStatus;
			info.DestinationLogisticUnitIdentKey = dstLuid;
			info.DestinationLocationCode = data.WeighingRoom.Code;
			info.DestinationPmxWhsCode = info.SourcePmxWhsCode;
			info.DestinationQualityStatusCode = info.SourceQualityStatusCode;

			//Catch weight
			info.Uom2 = null;
			info.QuantityUom2 = null;
		}

		/// <summary>
		/// Processes the weigh.
		/// </summary>
		/// <param name="methodHelper">The method helper.</param>
		/// <param name="weighData"></param>
		/// <param name="weight"></param>
		/// <param name="targetLuid"></param>
		/// <param name="deviceInstanceId"></param>
		private static void ProcessWeigh(WorkflowHelper methodHelper, WeighData weighData, double weight, int? targetLuid, string deviceInstanceId)
		{
			try
			{
				if (weight == 0)
				{
					return;
				}
				bool needsReceipt = weight > weighData.QtyAvailable;

				int luid = -1;

				while (true)
				{

					try
					{

						luid = methodHelper.SboProviderService.InvokeMethodWithDbConnection(needsReceipt, false, null, null,  // Use SBO when theoQtyCorrection is needed
							   delegate (PmxDbConnection dbConn, object[] parameters)
							   {
								   int dstLuid = MoveStockFromInputLocationToWeightRoom(weighData, weight, targetLuid, needsReceipt, dbConn);

								   weighData.SelectedLuid = null;

								   return dstLuid;
							   });

						break;
					}
					catch (Exception ex)
					{
						s_log.Error(ex.ToString());
						if (!(ex is ProdumexException) || ((ProdumexException)ex).Error == null)
						{
							ShowMessageScreenForException(methodHelper, "InitialErrorKey", methodHelper.InitialErrorKey,
									"MessageKey", KnownMessageKeys.MSG_ERROR_INTERNAL_ERROR,
									"MessageAddendum", ex.Message,
									"PmxError", Produmex.Sbo.Logex.Data.ProdumexErrors.UnexpectedError(s_log, ex),
									"ShowButton", true);
						}
						else
						{
							ShowMessageScreenForException(methodHelper, "InitialErrorKey", methodHelper.InitialErrorKey,
								"MessageKey", KnownMessageKeys.MSG_ERROR_INTERNAL_ERROR,
								"MessageAddendum", ex.Message,
								"PmxError", ((ProdumexException)ex).Error,
								"ShowButton", true);
						}



						//retry?
						//ask user if he wants to retry
						methodHelper.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IDecisionScreen),
							methodHelper.CultureInfo, BuildParamCollection(
								"TitleKey", "MSG_TITLE_RETRY_UPDATE",
								"MessageKey", "MSG_QUESTION_AN_ERROR_OCCURED_WANT_RETRY"));
						methodHelper.Msg = methodHelper.WaitForMessage();

						if (methodHelper.Msg.Name.ToLower().EndsWith(".no"))
						{
							break;
						}
					}
				}

				bool askReprint = false;

				weighData.QtyAvailable = Math.Max(0, weighData.QtyAvailable - weight);
				weighData.QtyToDo = Math.Min(Math.Max(0, weighData.QtyToDo - weight), weighData.QtyPlanned);

				// Print LABEL
				if (weighData.SelectedLuid != luid)
				{
					//print the label
					bool proceed = false;
					while (!proceed)
					{
						int numberOfReports = 0;

						try
						{
							//trigger event
							numberOfReports = methodHelper.SboProviderService.InvokeMethodWithDbConnection(false, false, null, null, delegate (PmxDbConnection conn, object[] parameters)
							{
								PmxPrintReportProvider printProv = new PmxPrintReportProvider(conn);
								return printProv.PrintReportsForEvent(PmxPrintReportEventType.WarehouseCreatedLogisticUnit, weighData.WeighingRoom.Code, deviceInstanceId, luid, out askReprint);
							});
						}
						catch (Produmex.Foundation.ProdumexException ex)
						{
							s_log.Error(ex.ToString());

							ShowMessageScreenForException(methodHelper, "InitialErrorKey", methodHelper.InitialErrorKey,
									"MessageKey", KnownMessageKeys.MSG_ERROR_INTERNAL_ERROR,
									"MessageAddendum", ex.Message,
									"PmxError", ex.Error,
									"ShowButton", true);
						}
						catch (Exception ex)
						{
							s_log.Error(ex.ToString());

							ShowMessageScreenForException(methodHelper, "InitialErrorKey", methodHelper.InitialErrorKey,
									"MessageKey", KnownMessageKeys.MSG_ERROR_INTERNAL_ERROR,
									"MessageAddendum", ex.Message,
									"PmxError", Produmex.Sbo.Logex.Data.ProdumexErrors.UnexpectedError(s_log, ex),
									"ShowButton", true);
						}

						if (numberOfReports > 0 && askReprint)
						{
							methodHelper.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IDecisionScreen),
								methodHelper.CultureInfo, BuildParamCollection(
									"MessageKey", "MSG_QUESTION_REPRINT_LABEL",
									"TitleKey", "MSG_TITLE_REPRINT_LABEL"));

							methodHelper.Msg = methodHelper.WaitForMessage();

							if (methodHelper.Msg.Name.EndsWith(".No"))
							{
								proceed = true;
							}
						}
						else
						{
							proceed = true;
						}
					}
				}

				weighData.SelectedLuid = luid;
			}
			catch (ProdumexException ex)
			{
				s_log.Error(ex.ToString());

				ShowMessageScreenForException(methodHelper, "PmxError", ex.Error, "ShowButton", true);
				methodHelper.InitialErrorKey = null;
			}
			catch (Exception ex)
			{
				s_log.Error(ex.ToString());

				ShowMessageScreenForException(methodHelper, "MessageKey", ExceptionHelper.Format(ex, methodHelper.CultureInfo),
						"NoTranslationOfMessageKey", true,
						"IsError", true,
						"ShowButton", true);
				methodHelper.InitialErrorKey = null;
			}
		}

		private static GotoResult GetWeighOrderLineData(WorkflowHelper methodHelper, string weighingLocation, string itemCode, int? weighOrderKey, WeighData weighData, GotoResult backGotoResult)
		{
			//adapt weigh order data
			methodHelper.SboProviderService.InvokeMethodWithDbConnection<object>(true, false, null, null,
		   delegate (PmxDbConnection conn, object[] parameters)
		   {

			   PmxProductionOrderProvider poProv = new PmxProductionOrderProvider((SboConnection)conn);
			   PmxWeighOrderProvider woProv = new PmxWeighOrderProvider(conn);

			   PmxWeighOrder wOrder = woProv.GetBO(weighOrderKey.Value);
			   PmxProductionOrder pOrder = poProv.GetBO(wOrder.BaseEntry.Value);

			   foreach (PmxWeighOrderLine line in wOrder.Lines)
			   {
				   //call stored procedure
				   if (line.OpenQuantity == line.Quantity) // oly execute sp when no weighin was performed yet
				   {
					   s_log.Debug("Start stored procedure potency recalculation ");
					   bool isHana = (methodHelper.SboProviderService.GetDbTool().DbServerType == DbServerType.HANA);
					   string qry = "";

					   DbQueryBuilder qryBuilder = methodHelper.SboProviderService.GetDbTool().GetQueryBuilder();
					   if (isHana)
					   {						 
						   qryBuilder.AppendFormat("CALL {0} ({1}, {2})",						  
							   qryBuilder.AsIdentifier(Definitions.STORED_PROCEDURE),
							   SqlCommandHelper.FormatColumnValue(wOrder.BaseEntry.Value, CultureInfo.InvariantCulture),
							   SqlCommandHelper.FormatColumnValue(line.ItemCode, CultureInfo.InvariantCulture));	
					   }
					   else {						  
						   qryBuilder.AppendFormat("EXEC {0} {1}, {2}",
							   qryBuilder.AsIdentifier(Definitions.STORED_PROCEDURE),
							   SqlCommandHelper.FormatColumnValue(wOrder.BaseEntry.Value, CultureInfo.InvariantCulture),
							   SqlCommandHelper.FormatColumnValue(line.ItemCode, CultureInfo.InvariantCulture));
					   }


					   qry = qryBuilder.ToString();
					   DataSet dsRecal = methodHelper.SboProviderService.RunView(false, null, null, qry);
					   s_log.Debug("Stop stored procedure potency recalculation ");

					   if (GetNumberOfRows(dsRecal) == 0)
					   {
						   s_log.Debug(String.Format("stored procedure potency recalculation gives no result for itemcode: {0} , production order docEntry:{1} ", line.BaseEntry, line.ItemCode));
					   }
					   else
					   {
						   DataRow row = dsRecal.Tables[0].Rows[0];
						   line.Quantity = Convert.ToDouble(row[Definitions.COLUMN_NAME]);
						   line.OpenQuantity = Convert.ToDouble(row[Definitions.COLUMN_NAME]);

						   pOrder.Lines[line.BaseLine.Value].BaseQuantity = Convert.ToDouble(row[Definitions.COLUMN_NAME]);
						   pOrder.Lines[line.BaseLine.Value].PlannedQuantity = Convert.ToDouble(row[Definitions.COLUMN_NAME]);
					   }

				   }
			   }

			   poProv.UpdateBO(pOrder, false, false);
			   woProv.UpdateBO(wOrder, false, false);

			   return null;
		   });
			

			string query = BuildQuery.GetWeighOrderLines(weighingLocation, weighOrderKey, itemCode, methodHelper.SboProviderService.GetDbTool());
			DataSet dsLines = methodHelper.SboProviderService.RunView(false, null, null, query);

			methodHelper.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.Weighing.ISelectWeighOrderLineScreen),
			   methodHelper.CultureInfo, BuildParamCollection(
				"InitialErrorKey", methodHelper.InitialErrorKey,
				"TitleKey", "MSG_TITLE_SELECT_ITEM_TO_WEIGH",
				"WeighOrderLineDS", dsLines
				));
			methodHelper.Msg = methodHelper.WaitForMessage();

			methodHelper.InitialErrorKey = null;

			if (methodHelper.Msg.Name.EndsWith(".WeighOrderLineSelected"))
			{
				int weighOrderLineKey = ExtractParameter<int>(methodHelper.Msg.Parameters, "weighOrderLineKey");

				string filter = WeighOrderLineDataSet.Columns.INTERNAL_KEY.Name + " = " + weighOrderLineKey;

				DataSet ds = dsLines.Clone();
				foreach (DataRow dr in dsLines.Tables[0].Select(filter))
				{
					ds.Tables[0].Rows.Add(dr.ItemArray);
				}

				DataRow row = ds.Tables[0].Rows[0];

				GetWeighOrderLineData_FillWeighData(weighData, row);

				weighData.SelectedLuid = null;
				weighData.TargetLuidDS = null;
				weighData.ScaleData = null;
				weighData.ScaleCode = null;
				weighData.ScaleCoefficient = null;

				return GotoResult.ScanSSCC;
			}

			if (methodHelper.Msg.Name.EndsWith(".RefreshRequested"))
			{
				return GotoResult.SelectWeighOrderLine;
			}

			return backGotoResult;
		}

		private static void GetWeighOrderLineData_FillWeighData(WeighData weighData, DataRow row)
		{
			weighData.ProductionOrderDocNum = Convert.ToInt32(row[WeighOrderLineDataSet.Columns.PROD_ORDER_NUM.Name]);
			weighData.ProductionOrderDocEntry = Convert.ToInt32(row[WeighOrderLineDataSet.Columns.PROD_ORDER_DOC_ENTRY.Name]);
			weighData.ProductionOrderLineNum = Convert.ToInt32(row[WeighOrderLineDataSet.Columns.PROD_ORDER_LINE_NUM.Name]);
			weighData.WeighOrderKey = Convert.ToInt32(row[WeighOrderLineDataSet.Columns.ORDER_NUM.Name]);
			weighData.WeighOrderLineNum = Convert.ToInt32(row[WeighOrderLineDataSet.Columns.ORDER_LINE_NUM.Name]);
			weighData.WeighOrderLineKey = Convert.ToInt32(row[WeighOrderLineDataSet.Columns.INTERNAL_KEY.Name]);
			weighData.ProductionOrderItemCode = Convert.ToString(row[WeighOrderLineDataSet.Columns.PROD_ITEM_CODE.Name]);
			weighData.ProductionOrderItemDescription = Convert.ToString(row[WeighOrderLineDataSet.Columns.PROD_ITEM_DESCR.Name]);
			weighData.LineItemCode = Convert.ToString(row[WeighOrderLineDataSet.Columns.LINE_ITEM_CODE.Name]);
			weighData.LineItemDescription = Convert.ToString(row[WeighOrderLineDataSet.Columns.LINE_ITEM_DESCR.Name]);
			weighData.LineItemName = Convert.ToString(row[WeighOrderLineDataSet.Columns.LINE_ITEM_NAME.Name]);
		}

		private static GotoResult GetWeighableItem(WorkflowHelper methodHelper, string weighingLocation, out string itemCode, GotoResult backGotoResult)
		{
			itemCode = null;

			string query = BuildQuery.GetWeighableItems(weighingLocation, methodHelper.SboProviderService.GetDbTool());
			DataSet dsItems = methodHelper.SboProviderService.RunView(false, null, null, query);

			if (GetNumberOfRows(dsItems) == 0)
			{
				methodHelper.InitialErrorKey = "MSG_ERROR_NO_WEIGHABLE_ITEMS_FOR_THE_WEIGHING_ROOM";
				return GotoResult.SelectWeighingRoom;
			}

			methodHelper.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.ISelectProductScreen),
			   methodHelper.CultureInfo, BuildParamCollection(
				"InitialErrorKey", methodHelper.InitialErrorKey,
				"TitleKey", "MSG_TITLE_SELECT_ITEM",
				"ProductDS", dsItems
				));
			methodHelper.Msg = methodHelper.WaitForMessage();

			methodHelper.InitialErrorKey = null;

			if (methodHelper.Msg.Name.EndsWith(".ProductSelected"))
			{
				itemCode = ExtractParameter<string>(methodHelper.Msg.Parameters, "itemCode");
				return GotoResult.SelectWeighOrderLine;
			}

			return backGotoResult;
		}

		private static GotoResult GetOrderToWeigh(WorkflowHelper methodHelper, string weighingLocation, out int? weighOrderKey, GotoResult backGotoResult)
		{
			weighOrderKey = null;

			DataSet dsWeighOrder = (DataSet)methodHelper.SboProviderService.InvokeMethodWithDbConnection<object>(false, false, null, null,
				delegate (PmxDbConnection conn, object[] parameters)
				{
					string query = BuildQuery.GetWeighOrders(weighingLocation, conn.DbTool);

					ISboRecordset rs = SboRecordsetHelper.RunQuery(s_log, query, conn);
					return SboRecordsetHelper.ConvertRecordsetToDataSet(s_log, rs);
				});

			if (GetNumberOfRows(dsWeighOrder) == 0)
			{
				methodHelper.InitialErrorKey = "MSG_ERROR_NO_WEIGHABLE_ORDERS_FOR_THE_WEIGHING_ROOM";
				return GotoResult.SelectWeighingRoom;
			}

			methodHelper.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.Weighing.ISelectWeighOrderScreen),
				methodHelper.CultureInfo, BuildParamCollection(
				"InitialErrorKey", methodHelper.InitialErrorKey,
				"TitleKey", "MSG_TITLE_SELECT_WEIGH_ORDER",
				"WeighOrderDS", dsWeighOrder));
			methodHelper.Msg = methodHelper.WaitForMessage();
			methodHelper.InitialErrorKey = null;

			if (methodHelper.Msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.Weighing.ISelectWeighOrderScreenReply.BackRequested")
			{
				return backGotoResult;
			}
			else if (methodHelper.Msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.Weighing.ISelectWeighOrderScreenReply.RefreshRequested")
			{
				return GotoResult.SelectOrderToWeigh;
			}
			else if (methodHelper.Msg.Name != "Produmex.Foundation.SlimScreen.Interfaces.Weighing.ISelectWeighOrderScreenReply.WeighOrderSelected")
			{
				methodHelper.InitialErrorKey = KnownMessageKeys.MSG_ERROR_OPERATION_NOT_SUPPORTED;
				return GotoResult.SelectOrderToWeigh;
			}
			else
			{
				weighOrderKey = ExtractParameter<int>(methodHelper.Msg.Parameters, "weighOrderKey");
				return GotoResult.SelectWeighOrderLine;
			}
		}

		private static GotoResult GetWeighingRoom(WorkflowHelper methodHelper, string warehouseCode, string deviceID, out string weighingLocation)
		{
			weighingLocation = null;

			string query = BuildQuery.GetWeighingLocation(warehouseCode, deviceID);
			DataSet dsWeighingLocations = methodHelper.SboProviderService.RunView(false, null, null, query);

			//let user select it
			methodHelper.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.ISelectLocationScreen),
			   methodHelper.CultureInfo, BuildParamCollection(
			   "InitialErrorKey", methodHelper.InitialErrorKey,
			   "TitleKey", "MSG_TITLE_SELECT_WEIGHING_ROOM",
			   "LocationDS", dsWeighingLocations,
			   "ShowStopButton", false));

			methodHelper.Msg = methodHelper.WaitForMessage();
			// Reset latest error
			methodHelper.InitialErrorKey = null;

			if (methodHelper.Msg.Name.EndsWith(".LocationSelected"))
			{
				weighingLocation = ExtractParameter<string>(methodHelper.Msg.Parameters, "locationCode");

				return GotoResult.CheckWeighStrategy;
			}
			return GotoResult.Exit;
		}

		private GotoResult CheckWeighStrategy(WorkflowHelper methodHelper, string weighingLocation, WeighData weighData)
		{
			weighData.WhsInfo = m_sboProviderService.InvokeMethodWithDbConnection(false, false, null, null, GetPmxWarehouseInfo, weighingLocation);

			string query = SboCommands.GetQueryGetAllScalesOfWeighingRooms(weighingLocation, true, methodHelper.CultureInfo);
			DataSet dsScales = methodHelper.SboProviderService.RunView(false, null, null, query);

			weighData.NumberOfScales = dsScales.Tables[0].Rows.Count;

			PmxWeighStrategy weighStrategy = methodHelper.SboProviderService.InvokeMethodWithDbConnection(false, false, null, null,
				delegate (PmxDbConnection conn, object[] parameters)
				{
					PmxOseWeighingRoomProvider weighRoomProvider = new PmxOseWeighingRoomProvider(conn);
					weighData.WeighingRoom = weighRoomProvider.GetBO(weighingLocation);
					PmxWeighStrategy strategy = weighData.WeighingRoom.DefaultWeighStrategy;
					return strategy;
				});

			if (weighStrategy == PmxWeighStrategy.ByItem)
			{
				return GotoResult.SelectItemToWeigh;
			}
			else
			{
				return GotoResult.SelectOrderToWeigh;
			}
		}

		private enum GotoResult
		{
			Exit,
			SelectOrderToWeigh,
			SelectItemToWeigh,
			CheckWeighStrategy,
			SelectWeighingRoom,
			SelectWeighOrderLine,
			WeighItem,
			ScanSSCC,
			SelectScale
		}

		private class WorkflowHelper
		{
			public ISboProviderService SboProviderService;
			public Action<Type, CultureInfo, ReadOnlyCollection<Parameter>> ShowScreen;
			public Func<Message> WaitForMessage;
			public CultureInfo CultureInfo;
			public Message Msg;
			public string InitialErrorKey;
		}

		private class WeighData
		{
			public PmxOseScale ScaleData;
			public string ScaleCode;
			public int NumberOfScales;
			public int WeighOrderLineNum;
			public int WeighOrderLineKey;
			public string LineItemCode;
			public int WeighOrderKey;
			public int ProductionOrderDocNum;
			public int ProductionOrderDocEntry;
			public string ProductionOrderItemCode;
			public string ProductionOrderItemDescription;
			public string LineItemDescription;
			public int? SelectedLuid;
			public DataSet TargetLuidDS;
			public string BatchNumber;
			public string BatchNumber2;
			public int? UomDecimals;
			public string InventoryUom;
			public int? InventoryUomDecimals;
			public string ScaleUom;
			public string StockSSCC;
			public DateTime? BestBeforeDate;
			public double QtyPlanned;
			public double QtyAvailable;
			public double QtyToDo;
			public PmxOseWeighingRoom WeighingRoom;
			public int? StockLuid;
			public int ProductionOrderLineNum;
			public double Tolerance;
			public double? ScaleCoefficient;
			public string QualityStatus;
			public int? ItriKey;
			public string LineItemName;
			public PmxOseWarehouseInfo WhsInfo;
		}

		#region HELPERS

		/// <summary>
		/// Gets the number of rows.
		/// </summary>
		/// <param name="ds">The ds.</param>
		/// <returns></returns>
		private static int GetNumberOfRows(DataSet ds)
		{
			if (ds != null &&
				ds.Tables.Count > 0)
			{
				return ds.Tables[0].Rows.Count;
			}

			return 0;
		}

		/// <summary>
		/// Executes the get PMX warehouse info.
		/// </summary>
		/// <param name="conn">The conn.</param>
		/// <param name="parameters">The parameters.</param>
		/// <returns></returns>
		private static PmxOseWarehouseInfo GetPmxWarehouseInfo(PmxDbConnection conn, object[] parameters)
		{
			PmxOseProvider oseProv = new PmxOseProvider(conn);
			return oseProv.GetPmxWarehouseInfo((string)parameters[0]);
		}

		#endregion

		#region BUILD QUERY

		/// <summary>
		/// Static class grouping our native queries
		/// </summary>
		private static class BuildQuery
		{
			internal static string GetWeighingLocation(string pmxWarehouseCode, string deviceID)
			{
				StringBuilder strBuilder = new StringBuilder();
				strBuilder.Append("SELECT ");

				SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.Code.NAME, LocationDataSet.Columns.CODE.Name);

				strBuilder.Append(", ");
				SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.Name.NAME, LocationDataSet.Columns.DESCRIPTION.Name);

				strBuilder.Append(", ");
				SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME, LocationDataSet.Columns.WAREHOUSECODE.Name);

				strBuilder.Append(", ");
				SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxOseWarehouseTable.NAME, PmxOseWarehouseTable.Columns.Name.NAME, LocationDataSet.Columns.WAREHOUSENAME.Name);

				strBuilder.Append(" FROM ");
				strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseStorageLocationGeneralTable.NAME));

				strBuilder.Append(" INNER JOIN ");
				strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseGeneralTable.NAME));
				strBuilder.Append(" ON ");
				SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.Code.NAME);
				strBuilder.Append(" = ");
				SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.Code.NAME);

				strBuilder.Append(" INNER JOIN ");
				strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseWarehouseTable.NAME));
				strBuilder.Append(" ON ");
				SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME);
				strBuilder.Append(" = ");
				SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseWarehouseTable.NAME, PmxOseWarehouseTable.Columns.Code.NAME);

				strBuilder.Append("\r\n INNER JOIN ");
				strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxWarehouseTable.NAME));
				strBuilder.Append(" ON ");
				SqlCommandHelper.AddTableColumnName(strBuilder, PmxWarehouseTable.NAME, PmxWarehouseTable.Columns.WarehouseCode.NAME);
				strBuilder.Append(" = ");
				SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseWarehouseTable.NAME, PmxOseWarehouseTable.Columns.SboWarehouseCode.NAME);

				strBuilder.Append("\r\n AND ");
				SqlCommandHelper.AddTableColumnName(strBuilder, PmxWarehouseTable.NAME, PmxWarehouseTable.Columns.IsManagedByProdumex.NAME);
				strBuilder.Append(" = ");
				strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));

				strBuilder.Append(" INNER JOIN ");
				strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseWeighingRoomTable.NAME));
				strBuilder.Append(" ON ");
				SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseWeighingRoomTable.NAME, PmxOseWeighingRoomTable.Columns.Code.NAME);
				strBuilder.Append(" = ");
				SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.Code.NAME);

				strBuilder.Append(" WHERE ");
				strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.NAME));
				strBuilder.Append(".");
				strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.Columns.IsActive.NAME));
				strBuilder.Append(" = ");
				strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));

				if (!string.IsNullOrEmpty(pmxWarehouseCode) || !string.IsNullOrEmpty(deviceID))
				{
					strBuilder.Append(" AND ( ");
					if (!string.IsNullOrEmpty(pmxWarehouseCode))
					{
						SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME);
						strBuilder.Append(" = ");
						strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(pmxWarehouseCode, CultureInfo.InvariantCulture));
					}
					if (!string.IsNullOrEmpty(deviceID))
					{
						if (!string.IsNullOrEmpty(pmxWarehouseCode))
							strBuilder.Append(" OR ");

						strBuilder.AppendFormat("{0} IN ( SELECT {1} FROM {2} WHERE {3} = {4} )",
							SqlCommandHelper.EscapeTableColumnName(PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME),
							SqlCommandHelper.EscapeTableColumnName(PmxThinClientWarehouseTable.NAME, PmxThinClientWarehouseTable.Columns.PmxWarehouseCode.NAME),
							SqlCommandHelper.EscapeTableName(PmxThinClientWarehouseTable.NAME),
							SqlCommandHelper.EscapeTableColumnName(PmxThinClientWarehouseTable.NAME, PmxThinClientWarehouseTable.Columns.ParentCode.NAME),
							SqlCommandHelper.FormatAndEscapeColumnValue(deviceID, CultureInfo.InvariantCulture));
					}
					strBuilder.Append(" )");
				}

				strBuilder.Append(" ORDER BY ");
				strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.NAME));
				strBuilder.Append(".");
				strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.Columns.Name.NAME));

				return strBuilder.ToString();
			}

			internal static string GetWeighOrders(string weighingLocation, DbTool dbTool)
			{
				DbQueryBuilder qb = dbTool.GetQueryBuilder();

				qb = qb.Select.
					EscTableColumnNameWithAlias(PmxWeighOrderTable.NAME, PmxWeighOrderTable.Columns.DocEntry.NAME, WeighOrderDataSet.Columns.INTERNAL_KEY.Name).Comma.
					EscTableColumnNameWithAlias(PmxProductionOrderTable.NAME, PmxProductionOrderTable.Columns.DocNum.NAME, WeighOrderDataSet.Columns.PROD_ORDER_NUM.Name).Comma.
					EscTableColumnNameWithAlias(PmxItemTable.NAME, PmxItemTable.Columns.CustomItemDescr.NAME, WeighOrderDataSet.Columns.ITEM_DESCR.Name).Comma.
					Append(" CAST( ").EscTableColumnName(PmxProductionOrderTable.NAME, PmxProductionOrderTable.Columns.DocNum.NAME).Append(" AS NVARCHAR) AS ").
					Append(WeighOrderDataSet.Columns.PROD_ORDER_NUM_AS_STRING.Name).Comma.
					Append(" CAST( ").EscTableColumnName(PmxWeighOrderTable.NAME, PmxWeighOrderTable.Columns.DocEntry.NAME).Append(" AS NVARCHAR) AS ").
					Append(WeighOrderDataSet.Columns.INTERNAL_KEY_AS_STRING.Name).
					CRLF.
				From.CRLF.
					EscTableName(PmxWeighOrderTable.NAME).
					CRLF.
				LeftJoin.
					EscTableName(PmxProductionOrderTable.NAME).On.
					EscTableColumnName(PmxProductionOrderTable.NAME, PmxProductionOrderTable.Columns.DocEntry.NAME).Eq.
					EscTableColumnName(PmxWeighOrderTable.NAME, PmxWeighOrderTable.Columns.BaseEntry.NAME).
					CRLF.
				LeftJoin.
					EscTableName(PmxItemTable.NAME).WithNoLock.On.
					EscTableColumnName(PmxItemTable.NAME, PmxItemTable.Columns.ItemCode.NAME).Eq.
					EscTableColumnName(PmxProductionOrderTable.NAME, PmxProductionOrderTable.Columns.ItemCode.NAME).
					CRLF.
				LeftJoin.
					EscTableName(PmxTroyLinkWeighOrderWeighRoomTable.NAME).WithNoLock.On.
					EscTableColumnName(PmxTroyLinkWeighOrderWeighRoomTable.NAME, PmxTroyLinkWeighOrderWeighRoomTable.Columns.WeighOrderDocEntry.NAME).Eq.
					EscTableColumnName(PmxWeighOrderTable.NAME, PmxWeighOrderTable.Columns.DocEntry.NAME).
					CRLF.
				Where.
					EscTableColumnName(PmxTroyLinkWeighOrderWeighRoomTable.NAME, PmxTroyLinkWeighOrderWeighRoomTable.Columns.WeighingRoomCode.NAME).Eq.
					AppendFormat(SqlCommandHelper.FormatAndEscapeColumnValue(weighingLocation, dbTool.CultureInfo)).
					And.
					CRLF.
					EscTableColumnName(PmxWeighOrderTable.NAME, PmxWeighOrderTable.Columns.DocStatus.NAME).Eq.
					Append(SqlCommandHelper.FormatAndEscapeColumnValue(PmxDocumentStatusColumnValues.Open.VALUE, dbTool.CultureInfo));

				return qb.ToString();
			}

			internal static string GetWeighOrderLines(string weighingLocation, int? weighOrder, string itemCode, DbTool dbTool)
			{
				const string PROD_ORDER_ITEM_PREFIX = "P";
				const string LINE_ITEM_PREFIX = "POLI";

				DbQueryBuilder qb = dbTool.GetQueryBuilder();

				qb = qb.Select.
					EscTableColumnNameWithAlias(PROD_ORDER_ITEM_PREFIX + PmxItemTable.NAME, PmxItemTable.Columns.ItemCode.NAME, WeighOrderLineDataSet.Columns.PROD_ITEM_CODE.Name).Comma.
					EscTableColumnNameWithAlias(PROD_ORDER_ITEM_PREFIX + PmxItemTable.NAME, PmxItemTable.Columns.CustomItemDescr.NAME, WeighOrderLineDataSet.Columns.PROD_ITEM_DESCR.Name).Comma.
					EscTableColumnNameWithAlias(LINE_ITEM_PREFIX + PmxItemTable.NAME, PmxItemTable.Columns.ItemCode.NAME, WeighOrderLineDataSet.Columns.LINE_ITEM_CODE.Name).Comma.
					EscTableColumnNameWithAlias(LINE_ITEM_PREFIX + PmxItemTable.NAME, PmxItemTable.Columns.CustomItemDescr.NAME, WeighOrderLineDataSet.Columns.LINE_ITEM_DESCR.Name).Comma.
					EscTableColumnNameWithAlias(LINE_ITEM_PREFIX + PmxItemTable.NAME, PmxItemTable.Columns.ItemName.NAME, WeighOrderLineDataSet.Columns.LINE_ITEM_NAME.Name).Comma.
					EscTableColumnNameWithAlias(PmxProductionOrderTable.NAME, PmxProductionOrderTable.Columns.DocEntry.NAME, WeighOrderLineDataSet.Columns.PROD_ORDER_DOC_ENTRY.Name).Comma.
					EscTableColumnNameWithAlias(PmxProductionOrderTable.NAME, PmxProductionOrderTable.Columns.DocNum.NAME, WeighOrderLineDataSet.Columns.PROD_ORDER_NUM.Name).Comma.
					EscTableColumnNameWithAlias(PmxProductionOrderLineTable.NAME, PmxProductionOrderLineTable.Columns.LineNum.NAME, WeighOrderLineDataSet.Columns.PROD_ORDER_LINE_NUM.Name).Comma.
					Append(" CAST( ").EscTableColumnName(PmxProductionOrderTable.NAME, PmxProductionOrderTable.Columns.DocNum.NAME).Append(" AS NVARCHAR) AS ").
					Append(WeighOrderLineDataSet.Columns.PROD_ORDER_NUM_AS_STRING.Name).Comma.
					EscTableColumnNameWithAlias(PmxWeighOrderLineTable.NAME, PmxWeighOrderLineTable.Columns.DocEntry.NAME, WeighOrderLineDataSet.Columns.ORDER_NUM.Name).Comma.
					Append(" CAST( ").EscTableColumnName(PmxWeighOrderLineTable.NAME, PmxWeighOrderLineTable.Columns.DocEntry.NAME).Append(" AS NVARCHAR) AS ").
					Append(WeighOrderLineDataSet.Columns.ORDER_NUM_AS_STRING.Name).Comma.
					EscTableColumnNameWithAlias(PmxWeighOrderLineTable.NAME, PmxWeighOrderLineTable.Columns.InternalKey.NAME, WeighOrderLineDataSet.Columns.INTERNAL_KEY.Name).Comma.
					Append(" CAST( ").EscTableColumnName(PmxWeighOrderLineTable.NAME, PmxWeighOrderLineTable.Columns.InternalKey.NAME).Append(" AS NVARCHAR) AS ").
					Append(WeighOrderLineDataSet.Columns.INTERNAL_KEY_AS_STRING.Name).Comma.
					EscTableColumnNameWithAlias(PmxWeighOrderLineTable.NAME, PmxWeighOrderLineTable.Columns.LineNum.NAME, WeighOrderLineDataSet.Columns.ORDER_LINE_NUM.Name).Comma.
					Append(" CAST( ").EscTableColumnName(PmxWeighOrderLineTable.NAME, PmxWeighOrderLineTable.Columns.LineNum.NAME).Append(" + 1 ").Append(" AS NVARCHAR) AS ").
					Append(WeighOrderLineDataSet.Columns.ORDER_LINE_AS_STRING.Name).Comma.
					EscTableColumnNameWithAlias(PmxWeighOrderLineTable.NAME, PmxWeighOrderLineTable.Columns.OpenQty.NAME, WeighOrderLineDataSet.Columns.QUANTITY_TO_WEIGH.Name).Comma.
					EscTableColumnNameWithAlias(PmxWeighOrderLineTable.NAME, PmxWeighOrderLineTable.Columns.Quantity.NAME, WeighOrderLineDataSet.Columns.TOTAL_QUANTITY.Name).Comma.
					EscTableColumnNameWithAlias(LINE_ITEM_PREFIX + PmxItemTable.NAME, PmxItemTable.Columns.InventoryUom.NAME, WeighOrderLineDataSet.Columns.UOM.Name).Comma.
					EscTableColumnNameWithAlias(LINE_ITEM_PREFIX + PmxItemTable.NAME, PmxItemTable.Columns.UomDecimals.NAME, WeighOrderLineDataSet.Columns.DECIMALS.Name).
					CRLF.
				From.CRLF.
					EscTableName(PmxWeighOrderLineTable.NAME).
					CRLF.
				LeftJoin.
					EscTableName(PmxWeighOrderTable.NAME).On.
					EscTableColumnName(PmxWeighOrderLineTable.NAME, PmxWeighOrderLineTable.Columns.DocEntry.NAME).Eq.
					EscTableColumnName(PmxWeighOrderTable.NAME, PmxWeighOrderTable.Columns.DocEntry.NAME).
					CRLF.
				LeftJoin.
					EscTableName(PmxProductionOrderTable.NAME).On.
					EscTableColumnName(PmxProductionOrderTable.NAME, PmxProductionOrderTable.Columns.DocEntry.NAME).Eq.
					EscTableColumnName(PmxWeighOrderTable.NAME, PmxWeighOrderTable.Columns.BaseEntry.NAME).
					CRLF.
				LeftJoin.
					EscTableName(PmxProductionOrderLineTable.NAME).On.
					EscTableColumnName(PmxProductionOrderTable.NAME, PmxProductionOrderTable.Columns.DocEntry.NAME).Eq.
					EscTableColumnName(PmxProductionOrderLineTable.NAME, PmxProductionOrderLineTable.Columns.DocEntry.NAME).
					And.
					EscTableColumnName(PmxWeighOrderLineTable.NAME, PmxWeighOrderLineTable.Columns.BaseLine.NAME).Eq.
					EscTableColumnName(PmxProductionOrderLineTable.NAME, PmxProductionOrderLineTable.Columns.LineNum.NAME).
					CRLF.
				LeftJoin.
					EscTableNameWithAlias(PmxItemTable.NAME, PROD_ORDER_ITEM_PREFIX + PmxItemTable.NAME).WithNoLock.On.
					EscTableColumnName(PROD_ORDER_ITEM_PREFIX + PmxItemTable.NAME, PmxItemTable.Columns.ItemCode.NAME).Eq.
					EscTableColumnName(PmxProductionOrderTable.NAME, PmxProductionOrderTable.Columns.ItemCode.NAME).
					CRLF.
				LeftJoin.
					EscTableNameWithAlias(PmxItemTable.NAME, LINE_ITEM_PREFIX + PmxItemTable.NAME).WithNoLock.On.
					EscTableColumnName(LINE_ITEM_PREFIX + PmxItemTable.NAME, PmxItemTable.Columns.ItemCode.NAME).Eq.
					EscTableColumnName(PmxProductionOrderLineTable.NAME, PmxProductionOrderLineTable.Columns.ItemCode.NAME).
					CRLF.				
				LeftJoin.
					EscTableName(PmxTroyLinkWeighOrderWeighRoomTable.NAME).On.
					EscTableColumnName(PmxTroyLinkWeighOrderWeighRoomTable.NAME, PmxTroyLinkWeighOrderWeighRoomTable.Columns.WeighOrderDocEntry.NAME).Eq.
					EscTableColumnName(PmxWeighOrderTable.NAME, PmxWeighOrderTable.Columns.DocEntry.NAME).
					CRLF.
				Where.
					CRLF.
					EscTableColumnName(PmxWeighOrderTable.NAME, PmxWeighOrderTable.Columns.DocStatus.NAME).Eq.
					Append(SqlCommandHelper.FormatAndEscapeColumnValue(PmxDocumentStatusColumnValues.Open.VALUE, dbTool.CultureInfo)).
					And.
					CRLF.
					EscTableColumnName(PmxWeighOrderLineTable.NAME, PmxWeighOrderLineTable.Columns.LineStatus.NAME).Eq.
					Append(SqlCommandHelper.FormatAndEscapeColumnValue(PmxDocumentLineStatusColumnValues.Open.VALUE, dbTool.CultureInfo)).
					And.
					CRLF.
					EscTableColumnName(PmxTroyLinkWeighOrderWeighRoomTable.NAME, PmxTroyLinkWeighOrderWeighRoomTable.Columns.WeighingRoomCode.NAME).Eq.
					AppendFormat(SqlCommandHelper.FormatAndEscapeColumnValue(weighingLocation, dbTool.CultureInfo));

				if (!string.IsNullOrWhiteSpace(itemCode))
				{
					qb = qb.And.
						EscTableColumnName(PmxWeighOrderLineTable.NAME, PmxWeighOrderLineTable.Columns.ItemCode.NAME).Eq.
						AppendFormat(SqlCommandHelper.FormatAndEscapeColumnValue(itemCode, dbTool.CultureInfo));
				}

				if (weighOrder.HasValue)
				{
					qb = qb.And.
						EscTableColumnName(PmxWeighOrderLineTable.NAME, PmxWeighOrderLineTable.Columns.DocEntry.NAME).Eq.
						AppendFormat(SqlCommandHelper.FormatAndEscapeColumnValue(weighOrder.Value, dbTool.CultureInfo));
				}

				return qb.ToString();
			}

			internal static string GetWeighableItems(string weighingLocation, DbTool dbTool)
			{
				DbQueryBuilder qb = dbTool.GetQueryBuilder();

				qb = qb.Select.Distinct.
					EscTableColumnNameWithAlias(PmxItemTable.NAME, PmxItemTable.Columns.ItemCode.NAME, ProductDataSet.Columns.PRODUCT_CODE.Name).Comma.
					EscTableColumnNameWithAlias(PmxItemTable.NAME, PmxItemTable.Columns.ItemName.NAME, ProductDataSet.Columns.PRODUCT_DESCRIPTION.Name).Comma.
					EscTableColumnNameWithAlias(PmxItemTable.NAME, PmxItemTable.Columns.CodeBars.NAME, ProductDataSet.Columns.GTIN.Name).
					CRLF.
				From.CRLF.
					EscTableName(PmxWeighOrderLineTable.NAME).
					CRLF.
				LeftJoin.
					EscTableName(PmxWeighOrderTable.NAME).On.
					EscTableColumnName(PmxWeighOrderLineTable.NAME, PmxWeighOrderLineTable.Columns.DocEntry.NAME).Eq.
					EscTableColumnName(PmxWeighOrderTable.NAME, PmxWeighOrderTable.Columns.DocEntry.NAME).
					CRLF.
				LeftJoin.
					EscTableName(PmxProductionOrderTable.NAME).On.
					EscTableColumnName(PmxProductionOrderTable.NAME, PmxProductionOrderTable.Columns.DocEntry.NAME).Eq.
					EscTableColumnName(PmxWeighOrderTable.NAME, PmxWeighOrderTable.Columns.BaseEntry.NAME).
					CRLF.And.
					EscTableColumnName(PmxWeighOrderTable.NAME, PmxWeighOrderTable.Columns.BaseType.NAME).Eq.
					Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboObjectTypes.ProductionOrder, dbTool.CultureInfo)).
					CRLF.
				LeftJoin.
					EscTableName(PmxProductionOrderLineTable.NAME).On.
					EscTableColumnName(PmxProductionOrderTable.NAME, PmxProductionOrderTable.Columns.DocEntry.NAME).Eq.
					EscTableColumnName(PmxProductionOrderLineTable.NAME, PmxProductionOrderLineTable.Columns.DocEntry.NAME).
					CRLF.And.
					EscTableColumnName(PmxWeighOrderLineTable.NAME, PmxWeighOrderLineTable.Columns.BaseLine.NAME).Eq.
					EscTableColumnName(PmxProductionOrderLineTable.NAME, PmxProductionOrderLineTable.Columns.LineNum.NAME).
					CRLF.
				LeftJoin.
					EscTableName(PmxItemTable.NAME).WithNoLock.On.
					EscTableColumnName(PmxItemTable.NAME, PmxItemTable.Columns.ItemCode.NAME).Eq.
					EscTableColumnName(PmxProductionOrderLineTable.NAME, PmxProductionOrderLineTable.Columns.ItemCode.NAME).
					CRLF.
				Where.
					CRLF.
					EscTableColumnName(PmxWeighOrderTable.NAME, PmxWeighOrderTable.Columns.DocStatus.NAME).Eq.
					Append(SqlCommandHelper.FormatAndEscapeColumnValue(PmxDocumentStatusColumnValues.Open.VALUE, dbTool.CultureInfo)).
					And.
					CRLF.
					EscTableColumnName(PmxWeighOrderLineTable.NAME, PmxWeighOrderLineTable.Columns.LineStatus.NAME).Eq.
					Append(SqlCommandHelper.FormatAndEscapeColumnValue(PmxDocumentLineStatusColumnValues.Open.VALUE, dbTool.CultureInfo)).
					And.
					CRLF.
					EscTableColumnName(PmxWeighOrderTable.NAME, PmxWeighOrderTable.Columns.WeighingRoomCode.NAME).Eq.
					AppendFormat(SqlCommandHelper.FormatAndEscapeColumnValue(weighingLocation, dbTool.CultureInfo));

				return qb.ToString();
			}

			/// <summary>
			/// Gets the stocks for item.
			/// </summary>
			/// <param name="weighOrderNumber">The weigh order number.</param>
			/// <param name="inputLocationCode">The input location code.</param>
			/// <param name="itemCode">The item code.</param>
			/// <param name="weighOrderLineKey"></param>
			/// <param name="dbTool">The database tool.</param>
			/// <returns></returns>
			internal static string GetStocksForItem(int weighOrderNumber, string inputLocationCode, string itemCode, int weighOrderLineKey, DbTool dbTool)
			{
                bool isHana = dbTool.DbServerType == DbServerType.HANA;

                DbQueryBuilder qryBuilder = dbTool.GetQueryBuilder();

                if (isHana)
                {
                    qryBuilder.Append("DO BEGIN \r\n");
                }

				qryBuilder.AppendFormat("SELECT {0}",
					SqlCommandHelper.EscapeTableColumnNameWithAlias(PmxWeighOrderLineTable.NAME, PmxWeighOrderLineTable.Columns.InternalKey.NAME, SelectItemToPrepareDataSet.Columns.LINE_NUM.Name));
				qryBuilder.AppendFormat(", {0}",
					SqlCommandHelper.EscapeTableColumnNameWithAlias(PmxItemTable.NAME, PmxItemTable.Columns.ItemCode.NAME, SelectItemToPrepareDataSet.Columns.ITEM_CODE.Name));
				qryBuilder.AppendFormat(", {0}",
					SqlCommandHelper.EscapeTableColumnNameWithAlias(PmxItemTable.NAME, PmxItemTable.Columns.ItemName.NAME, SelectItemToPrepareDataSet.Columns.ITEM_DESCR.Name));

				qryBuilder.AppendFormat(", " + qryBuilder.IFNULL + "({0}, {1}) AS {2} ",
				  SqlCommandHelper.EscapeTableColumnName(SboUomTable.NAME, SboUomTable.Columns.UomCode.NAME),
				  SqlCommandHelper.EscapeTableColumnName(SboItemTable.NAME, SboItemTable.Columns.InventoryUom.NAME),
				  SqlCommandHelper.EscapeColumnName(SelectItemToPrepareDataSet.Columns.UOM.Name));

				qryBuilder.AppendFormat(", {0}",
					SqlCommandHelper.EscapeTableColumnNameWithAlias(PmxItemTable.NAME, PmxItemTable.Columns.UomDecimals.NAME, SelectItemToPrepareDataSet.Columns.UOM_DECIMALS.Name));
				qryBuilder.AppendFormat(", {0}",
					SqlCommandHelper.EscapeTableColumnNameWithAlias(PmxWeighOrderLineTable.NAME, PmxWeighOrderLineTable.Columns.Quantity.NAME, SelectItemToPrepareDataSet.Columns.QTY_PLANNED.Name));
				qryBuilder.AppendFormat(", {0}",
					SqlCommandHelper.EscapeTableColumnNameWithAlias(PmxProductionOrderLineTable.NAME, PmxProductionOrderLineTable.Columns.QuantityTolerancePercentage.NAME, SelectItemToPrepareDataSet.Columns.TOLERANCE_PCT.Name));
				qryBuilder.AppendFormat(", {0}",
					SqlCommandHelper.EscapeTableColumnNameWithAlias(PmxWeighOrderLineTable.NAME, PmxWeighOrderLineTable.Columns.QuantityWeighed.NAME, SelectItemToPrepareDataSet.Columns.QTY_ISSUED.Name));
				qryBuilder.AppendFormat(", {0}",
					SqlCommandHelper.EscapeTableColumnNameWithAlias(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.ItemTransactionalInfoKey.NAME, SelectItemToPrepareDataSet.Columns.ITRI_KEY.Name));
				qryBuilder.AppendFormat(", {0}",
					SqlCommandHelper.EscapeTableColumnNameWithAlias(PmxItemTransactionalInfoTable.NAME, PmxItemTransactionalInfoTable.Columns.BatchNumber1.NAME, SelectItemToPrepareDataSet.Columns.BATCH_NUMBER.Name));
				qryBuilder.AppendFormat(", {0}",
					SqlCommandHelper.EscapeTableColumnNameWithAlias(PmxItemTransactionalInfoTable.NAME, PmxItemTransactionalInfoTable.Columns.BatchNumber2.NAME, SelectItemToPrepareDataSet.Columns.BATCH_NUMBER2.Name));
				qryBuilder.AppendFormat(", {0}",
					SqlCommandHelper.EscapeTableColumnNameWithAlias(PmxItemTransactionalInfoTable.NAME, PmxItemTransactionalInfoTable.Columns.BestBeforeDate.NAME, SelectItemToPrepareDataSet.Columns.BEST_BEFORE_DATE.Name));

				qryBuilder.AppendFormat(", {0}",
					SqlCommandHelper.EscapeTableColumnNameWithAlias(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.Quantity.NAME, SelectItemToPrepareDataSet.Columns.QUANTITY.Name));

				qryBuilder.AppendFormat(", {0}",
					SqlCommandHelper.EscapeTableColumnNameWithAlias(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.QualityStatusCode.NAME, SelectItemToPrepareDataSet.Columns.QUALITY_STATUS.Name));
				qryBuilder.AppendFormat(", {0}",
					SqlCommandHelper.EscapeTableColumnNameWithAlias(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.LogUnitIdentKey.NAME, SelectItemToPrepareDataSet.Columns.LUID.Name));
				qryBuilder.AppendFormat(", {0}",
					SqlCommandHelper.EscapeTableColumnNameWithAlias(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.StorLocCode.NAME, SelectItemToPrepareDataSet.Columns.STORAGE_LOCATION.Name));
				qryBuilder.AppendFormat(", {0}",
					SqlCommandHelper.EscapeTableColumnNameWithAlias(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.SSCC.NAME, SelectItemToPrepareDataSet.Columns.SSCC.Name));

				qryBuilder.AppendFormat("\r\n FROM  {0}",
					SqlCommandHelper.EscapeTableName(PmxWeighOrderTable.NAME));
				qryBuilder.AppendFormat("\r\n INNER JOIN {0}",
					SqlCommandHelper.EscapeTableName(PmxWeighOrderLineTable.NAME));
				qryBuilder.AppendFormat(" ON {0} = {1}",
					SqlCommandHelper.EscapeTableColumnName(PmxWeighOrderLineTable.NAME, PmxWeighOrderLineTable.Columns.DocEntry.NAME),
					SqlCommandHelper.EscapeTableColumnName(PmxWeighOrderTable.NAME, PmxWeighOrderTable.Columns.DocEntry.NAME));
				qryBuilder.AppendFormat("\r\n INNER JOIN {0}",
					SqlCommandHelper.EscapeTableName(PmxProductionOrderLineTable.NAME));
				qryBuilder.AppendFormat(" ON {0} = {1}",
					SqlCommandHelper.EscapeTableColumnName(PmxProductionOrderLineTable.NAME, PmxProductionOrderLineTable.Columns.LineNum.NAME),
					SqlCommandHelper.EscapeTableColumnName(PmxWeighOrderLineTable.NAME, PmxWeighOrderLineTable.Columns.BaseLine.NAME));
				qryBuilder.AppendFormat(" AND {0} = {1}",
					SqlCommandHelper.EscapeTableColumnName(PmxProductionOrderLineTable.NAME, PmxProductionOrderLineTable.Columns.DocEntry.NAME),
					SqlCommandHelper.EscapeTableColumnName(PmxWeighOrderTable.NAME, PmxWeighOrderTable.Columns.BaseEntry.NAME));

				qryBuilder.AppendFormat("\r\n INNER JOIN {0}",
					SqlCommandHelper.EscapeTableName(PmxItemTable.NAME));
				qryBuilder.Append(qryBuilder.WITH_NOLOCK);
				qryBuilder.AppendFormat(" ON {0} = {1}",
					SqlCommandHelper.EscapeTableColumnName(PmxItemTable.NAME, PmxItemTable.Columns.ItemCode.NAME),
					SqlCommandHelper.EscapeTableColumnName(PmxProductionOrderLineTable.NAME, PmxProductionOrderLineTable.Columns.ItemCode.NAME));

				qryBuilder.Append(" LEFT JOIN ");
				qryBuilder.Append(SqlCommandHelper.EscapeTableName(SboUomTable.NAME));
				qryBuilder.Append(" ON ");
				SqlCommandHelper.AddTableColumnName(qryBuilder, SboUomTable.NAME, SboUomTable.Columns.UomEntry.NAME);
				qryBuilder.Append(" = ");
				SqlCommandHelper.AddTableColumnName(qryBuilder, SboItemTable.NAME, SboItemTable.Columns.InventoryUoMEntry.NAME);
				qryBuilder.Append(" AND ");
				SqlCommandHelper.AddTableColumnName(qryBuilder, SboUomTable.NAME, SboUomTable.Columns.UomEntry.NAME);
				qryBuilder.Append(" <> -1 ");

				qryBuilder.AppendFormat("\r\n LEFT JOIN (  {0}",
					SqlCommandHelper.EscapeTableName(PmxInventoryTotalTable.NAME));
				{
					qryBuilder.AppendFormat("\r\n    LEFT JOIN {0}",
						SqlCommandHelper.EscapeTableName(PmxItemTransactionalInfoTable.NAME));
					qryBuilder.AppendFormat(" ON {0} = {1}",
						SqlCommandHelper.EscapeTableColumnName(PmxItemTransactionalInfoTable.NAME, PmxItemTransactionalInfoTable.Columns.InternalKey.NAME),
						SqlCommandHelper.EscapeTableColumnName(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.ItemTransactionalInfoKey.NAME));
				}
				qryBuilder.AppendFormat("\r\n ) ON {0} = {1}",
					SqlCommandHelper.EscapeTableColumnName(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.ItemCode.NAME),
					SqlCommandHelper.EscapeTableColumnName(PmxProductionOrderLineTable.NAME, PmxProductionOrderLineTable.Columns.ItemCode.NAME));
				qryBuilder.AppendFormat(" AND {0} = {1} ",
					SqlCommandHelper.EscapeTableColumnName(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.StorLocCode.NAME),
					SqlCommandHelper.FormatAndEscapeColumnValue(inputLocationCode, dbTool.CultureInfo));
				qryBuilder.AppendFormat(" AND ( {0} IS NULL OR {0} NOT IN ( SELECT {1} FROM {2} ) )",
					SqlCommandHelper.EscapeTableColumnName(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.LogUnitIdentKey.NAME),
					SqlCommandHelper.EscapeTableColumnName(PmxProductionOrderPreparedComponentTable.NAME, PmxProductionOrderPreparedComponentTable.Columns.LogUnitIdentKey.NAME),
					SqlCommandHelper.EscapeTableName(PmxProductionOrderPreparedComponentTable.NAME),
					SqlCommandHelper.EscapeTableColumnName(PmxProductionOrderPreparedComponentTable.NAME, PmxProductionOrderPreparedComponentTable.Columns.ProdOrderDocEntry.NAME),
					SqlCommandHelper.EscapeTableColumnName(PmxProductionOrderLineTable.NAME, PmxProductionOrderLineTable.Columns.DocEntry.NAME));

				qryBuilder.AppendFormat("\r\n WHERE {0} = {1}",
					SqlCommandHelper.EscapeTableColumnName(PmxWeighOrderLineTable.NAME, PmxWeighOrderLineTable.Columns.DocEntry.NAME),
					SqlCommandHelper.FormatAndEscapeColumnValue(weighOrderNumber, dbTool.CultureInfo));

				qryBuilder.AppendFormat(" AND {0} = {1}",
					SqlCommandHelper.EscapeTableColumnName(PmxItemTable.NAME, PmxItemTable.Columns.ItemCode.NAME),
					SqlCommandHelper.FormatAndEscapeColumnValue(itemCode, dbTool.CultureInfo));

				qryBuilder.AppendFormat("\r\n AND {0} = {1}",
					SqlCommandHelper.EscapeTableColumnName(PmxWeighOrderLineTable.NAME, PmxWeighOrderLineTable.Columns.InternalKey.NAME),
					SqlCommandHelper.FormatAndEscapeColumnValue(weighOrderLineKey, dbTool.CultureInfo));

				qryBuilder.AppendFormat("\r\n ORDER BY {0}",
					SqlCommandHelper.EscapeTableColumnName(PmxProductionOrderLineTable.NAME, PmxProductionOrderLineTable.Columns.LineNum.NAME));
				qryBuilder.AppendFormat(", {0};",
					SqlCommandHelper.EscapeTableColumnName(PmxItemTransactionalInfoTable.NAME, PmxItemTransactionalInfoTable.Columns.BestBeforeDate.NAME));

				qryBuilder.Append(GetPreparedQuantitites(weighOrderNumber, weighOrderLineKey, dbTool));

                if (isHana)
                {
                    qryBuilder.Append("\r\n END;");
                }


                return qryBuilder.ToString();
			}

			public static string GetPreparedQuantitites(int weighOrderNumber, int weighOrderLineKey, DbTool dbTool)
			{
				DbQueryBuilder qryBuilder = dbTool.GetQueryBuilder();

				// * Get already prepared quantities *
				qryBuilder.AppendFormat("\r\n\r\n SELECT {0}",
					SqlCommandHelper.EscapeTableColumnNameWithAlias(PmxProductionOrderPreparedComponentTable.NAME, PmxProductionOrderPreparedComponentTable.Columns.ProdBatchSeqNr.NAME, SelectItemToPrepareDataSet.Columns.PROD_BATCH_SEQ_NR.Name));
				qryBuilder.AppendFormat(", {0}",
					SqlCommandHelper.EscapeTableColumnNameWithAlias(PmxProductionOrderLineTable.NAME, PmxProductionOrderLineTable.Columns.LineNum.NAME, SelectItemToPrepareDataSet.Columns.LINE_NUM.Name));
				qryBuilder.AppendFormat(", {0}",
					SqlCommandHelper.EscapeTableColumnNameWithAlias(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.ItemCode.NAME, SelectItemToPrepareDataSet.Columns.ITEM_CODE.Name));
				qryBuilder.AppendFormat(", {0}",
					SqlCommandHelper.EscapeTableColumnNameWithAlias(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.Quantity.NAME, SelectItemToPrepareDataSet.Columns.QUANTITY.Name));
				qryBuilder.AppendFormat(", {0}",
					SqlCommandHelper.EscapeTableColumnNameWithAlias(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.LogUnitIdentKey.NAME, SelectItemToPrepareDataSet.Columns.LUID.Name));
				qryBuilder.AppendFormat(", {0}",
					SqlCommandHelper.EscapeTableColumnNameWithAlias(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.StorLocCode.NAME, SelectItemToPrepareDataSet.Columns.STORAGE_LOCATION.Name));
				qryBuilder.AppendFormat(", {0}",
					SqlCommandHelper.EscapeTableColumnNameWithAlias(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.SSCC.NAME, SelectItemToPrepareDataSet.Columns.SSCC.Name));
				qryBuilder.AppendFormat(", SUM( {0} ) OVER ( PARTITION BY {1} ) AS {2}",
					SqlCommandHelper.EscapeTableColumnName(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.Quantity.NAME),
					SqlCommandHelper.EscapeTableColumnName(PmxProductionOrderLineTable.NAME, PmxProductionOrderLineTable.Columns.LineNum.NAME),
					SqlCommandHelper.EscapeColumnName(SelectItemToPrepareDataSet.Columns.TOTAL_PREPARED_QTY.Name));

				qryBuilder.AppendFormat(", SUM( {0} ) OVER ( PARTITION BY {1}, {2} ) AS {3}",
					SqlCommandHelper.EscapeTableColumnName(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.Quantity.NAME),
					SqlCommandHelper.EscapeTableColumnName(PmxProductionOrderLineTable.NAME, PmxProductionOrderLineTable.Columns.LineNum.NAME),
					SqlCommandHelper.EscapeTableColumnName(PmxProductionOrderPreparedComponentTable.NAME, PmxProductionOrderPreparedComponentTable.Columns.ProdBatchSeqNr.NAME),
					SqlCommandHelper.EscapeColumnName(SelectItemToPrepareDataSet.Columns.TOTAL_PREPARED_QTY_BY_BATCH.Name));

				qryBuilder.AppendFormat(",  {0} ",
					 SqlCommandHelper.EscapeTableColumnNameWithAlias(PmxWeighOrderLineTable.NAME, PmxWeighOrderLineTable.Columns.Quantity.NAME, SelectItemToPrepareDataSet.Columns.QTY_PLANNED_BY_BATCH.Name));

				qryBuilder.AppendFormat("\r\n FROM  {0}",
					SqlCommandHelper.EscapeTableName(PmxWeighOrderTable.NAME));
				qryBuilder.AppendFormat("\r\n INNER JOIN {0}",
					SqlCommandHelper.EscapeTableName(PmxWeighOrderLineTable.NAME));
				qryBuilder.AppendFormat(" ON {0} = {1}",
					SqlCommandHelper.EscapeTableColumnName(PmxWeighOrderLineTable.NAME, PmxWeighOrderLineTable.Columns.DocEntry.NAME),
					SqlCommandHelper.EscapeTableColumnName(PmxWeighOrderTable.NAME, PmxWeighOrderTable.Columns.DocEntry.NAME));
				qryBuilder.AppendFormat("\r\n INNER JOIN {0}",
					SqlCommandHelper.EscapeTableName(PmxProductionOrderLineTable.NAME));
				qryBuilder.AppendFormat(" ON {0} = {1}",
					SqlCommandHelper.EscapeTableColumnName(PmxProductionOrderLineTable.NAME, PmxProductionOrderLineTable.Columns.LineNum.NAME),
					SqlCommandHelper.EscapeTableColumnName(PmxWeighOrderLineTable.NAME, PmxWeighOrderLineTable.Columns.BaseLine.NAME));
				qryBuilder.AppendFormat(" AND {0} = {1}",
					SqlCommandHelper.EscapeTableColumnName(PmxProductionOrderLineTable.NAME, PmxProductionOrderLineTable.Columns.DocEntry.NAME),
					SqlCommandHelper.EscapeTableColumnName(PmxWeighOrderTable.NAME, PmxWeighOrderTable.Columns.BaseEntry.NAME));

				qryBuilder.AppendFormat("\r\n INNER JOIN {0}",
					SqlCommandHelper.EscapeTableName(PmxProductionOrderTable.NAME));
				qryBuilder.AppendFormat(" ON {0} = {1}",
					SqlCommandHelper.EscapeTableColumnName(PmxProductionOrderTable.NAME, PmxProductionOrderTable.Columns.DocEntry.NAME),
					SqlCommandHelper.EscapeTableColumnName(PmxProductionOrderLineTable.NAME, PmxProductionOrderLineTable.Columns.DocEntry.NAME));

				qryBuilder.AppendFormat("\r\n INNER JOIN {0}",
					SqlCommandHelper.EscapeTableName(PmxProductionOrderPreparedComponentTable.NAME));
				qryBuilder.AppendFormat(" ON {0} = {1}",
					SqlCommandHelper.EscapeTableColumnName(PmxProductionOrderPreparedComponentTable.NAME, PmxProductionOrderPreparedComponentTable.Columns.ProdOrderDocEntry.NAME),
					SqlCommandHelper.EscapeTableColumnName(PmxProductionOrderLineTable.NAME, PmxProductionOrderLineTable.Columns.DocEntry.NAME));
				qryBuilder.AppendFormat(" AND {2}({0}, {1}) = {1}",
					SqlCommandHelper.EscapeTableColumnName(PmxProductionOrderPreparedComponentTable.NAME, PmxProductionOrderPreparedComponentTable.Columns.ProdOrderLineNum.NAME),
					SqlCommandHelper.EscapeTableColumnName(PmxProductionOrderLineTable.NAME, PmxProductionOrderLineTable.Columns.LineNum.NAME),
					qryBuilder.IFNULL);

				qryBuilder.AppendFormat("\r\n INNER JOIN {0}",
					SqlCommandHelper.EscapeTableName(PmxInventoryTotalTable.NAME));
				qryBuilder.AppendFormat(" ON {0} = {1}",
					SqlCommandHelper.EscapeTableColumnName(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.LogUnitIdentKey.NAME),
					SqlCommandHelper.EscapeTableColumnName(PmxProductionOrderPreparedComponentTable.NAME, PmxProductionOrderPreparedComponentTable.Columns.LogUnitIdentKey.NAME));
				qryBuilder.AppendFormat(" AND {0} = {1}",
					SqlCommandHelper.EscapeTableColumnName(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.ItemCode.NAME),
					SqlCommandHelper.EscapeTableColumnName(PmxProductionOrderLineTable.NAME, PmxProductionOrderLineTable.Columns.ItemCode.NAME));

				qryBuilder.AppendFormat("\r\n WHERE {0} = {1}",
					SqlCommandHelper.EscapeTableColumnName(PmxWeighOrderTable.NAME, PmxWeighOrderTable.Columns.DocEntry.NAME),
					SqlCommandHelper.FormatAndEscapeColumnValue(weighOrderNumber, dbTool.CultureInfo));

				qryBuilder.AppendFormat("\r\n AND {0} = {1}",
					SqlCommandHelper.EscapeTableColumnName(PmxWeighOrderLineTable.NAME, PmxWeighOrderLineTable.Columns.InternalKey.NAME),
					SqlCommandHelper.FormatAndEscapeColumnValue(weighOrderLineKey, dbTool.CultureInfo));

				qryBuilder.AppendFormat("\r\n ORDER BY {0};",
					SqlCommandHelper.EscapeTableColumnName(PmxProductionOrderLineTable.NAME, PmxProductionOrderLineTable.Columns.LineNum.NAME));

				return qryBuilder.ToString();
			}

			internal static string GetProductionOrderPreparedComponentWhereClause(int prodOrderDocEntry, int? prodBatchSeqNr, int luid, string itemCode, int? lineNum, PmxDbConnection dbConn)
			{
				StringBuilder strBuilder = new StringBuilder();

				strBuilder.AppendFormat("{0} = {1}",
					SqlCommandHelper.EscapeTableColumnName(PmxProductionOrderPreparedComponentTable.NAME, PmxProductionOrderPreparedComponentTable.Columns.ProdOrderDocEntry.NAME),
					SqlCommandHelper.FormatAndEscapeColumnValue(prodOrderDocEntry, dbConn.DbCultureInfo));
				strBuilder.AppendFormat(" AND {0} {1} {2}",
					SqlCommandHelper.EscapeTableColumnName(PmxProductionOrderPreparedComponentTable.NAME, PmxProductionOrderPreparedComponentTable.Columns.ProdBatchSeqNr.NAME),
					prodBatchSeqNr.HasValue ? "=" : "IS NULL",
					!prodBatchSeqNr.HasValue ? "" : SqlCommandHelper.FormatAndEscapeColumnValue(prodBatchSeqNr, dbConn.DbCultureInfo));
				strBuilder.AppendFormat(" AND {0} = {1}",
					SqlCommandHelper.EscapeTableColumnName(PmxProductionOrderPreparedComponentTable.NAME, PmxProductionOrderPreparedComponentTable.Columns.LogUnitIdentKey.NAME),
					SqlCommandHelper.FormatAndEscapeColumnValue(luid, dbConn.DbCultureInfo));
				strBuilder.AppendFormat(" AND {0} = {1}",
					SqlCommandHelper.EscapeTableColumnName(PmxProductionOrderPreparedComponentTable.NAME, PmxProductionOrderPreparedComponentTable.Columns.ItemCode.NAME),
					SqlCommandHelper.FormatAndEscapeColumnValue(itemCode, dbConn.DbCultureInfo));
				if (lineNum.HasValue)
				{
					strBuilder.AppendFormat(" AND {0} = {1}",
						SqlCommandHelper.EscapeTableColumnName(PmxProductionOrderPreparedComponentTable.NAME, PmxProductionOrderPreparedComponentTable.Columns.ProdOrderLineNum.NAME),
						SqlCommandHelper.FormatAndEscapeColumnValue(lineNum.Value, dbConn.DbCultureInfo));
				}

				return strBuilder.ToString();
			}
		}

		#endregion
	}
}
